"""This module implements the LiteraturePrior class which automates
some of the steps involved in starting a model around a set of
literature searches. Example:

.. code:: python

    lp = LiteraturePrior('some_disease', 'Some Disease',
                         'This is a self-updating model of Some Disease',
                         search_strings=['some disease'],
                         assembly_config_template='nf')
    estmts = lp.get_statements()
    model = lp.make_model(estmts, upload_to_s3=True)

"""
import tqdm
import logging
import datetime
from collections import defaultdict
from indra.util import batch_iter
from indra_db import get_db
from indra_db.util import distill_stmts
from indra_db.client.principal import get_raw_stmt_jsons_from_papers
from indra.databases import mesh_client
from indra.statements import stmts_from_json
from . import SearchTerm
from emmaa.model import EmmaaModel
from emmaa.statements import EmmaaStatement


logger = logging.getLogger(__name__)


class LiteraturePrior:
    def __init__(self, name, human_readable_name, description,
                 search_strings=None, mesh_ids=None,
                 assembly_config_template=None):
        """A class to construct a literture-based prior for an EMMAA model.

        Parameters
        ----------
        name : str
            The model name by which the model will be identified on S3.
        human_readable_name : str
            The human-readable display name for the model.
        description : str
            A human-readable description for the model.
        search_strings : list of str
            A list of search strings e.g., "diabetes" to find papers in the
            literature.
        mesh_ids : list of str
            A list of MeSH IDs that are used to search the literature as
            headings associated with papers.
        assembly_config_template : Optional[str]
            The name of another model from which the initial assembly
            configuration should be adopted.
        """
        self.name = name
        self.human_readable_name = human_readable_name
        self.description = description
        self.search_terms = \
            make_search_terms(search_strings, mesh_ids)
        if assembly_config_template:
            self.assembly_config = \
                self.get_config_from(assembly_config_template)
        else:
            self.assembly_config = {}

    def get_statements(self, mode='all', batch_size=100):
        """Return EMMAA Statements for this prior's literature set.

        Parameters
        ----------
        mode : 'all' or 'distilled'
            The 'distilled' mode makes sure that the "best", non-redundant
            set of raw statements are found across potentially redundant text
            contents and reader versions. The 'all' mode doesn't do such
            distillation but is significantly faster.
        batch_size : Optional[int]
            Determines how many PMIDs to fetch statements for in each
            iteration. Default: 100.

        Returns
        -------
        list of EmmaaStatement
            A list of EMMAA Statements corresponding to extractions from
            the subset of literature defined by this prior's search terms.
        """
        terms_to_pmids = \
            EmmaaModel.search_pubmed(search_terms=self.search_terms,
                                     date_limit=None)
        pmids_to_terms = defaultdict(list)
        for term, pmids in terms_to_pmids.items():
            for pmid in pmids:
                pmids_to_terms[pmid].append(term)
        pmids_to_terms = dict(pmids_to_terms)
        all_pmids = set(pmids_to_terms.keys())
        raw_statements_by_pmid = \
            get_raw_statements_for_pmids(all_pmids, mode=mode,
                                         batch_size=batch_size)
        timestamp = datetime.datetime.now()
        estmts = []
        for pmid, stmts in raw_statements_by_pmid.items():
            for stmt in stmts:
                estmts.append(EmmaaStatement(stmt, timestamp,
                                             pmids_to_terms[pmid]))
        return estmts

    def get_config_from(self, assembly_config_template):
        """Return assembly config given a template model's name.

        Parameters
        ----------
        assembly_config_template : str
            The name of a model whose assembly config should be adopted.

        Returns
        -------
        dict
            The assembly config of the given template model.
        """
        from emmaa.model import load_config_from_s3
        config = load_config_from_s3(assembly_config_template)
        return config.get('assembly')

    def make_config(self, upload_to_s3=False):
        """Return a config dict fot the model, optionally upload to S3.

        Parameters
        ----------
        upload_to_s3 : Optional[bool]
            If True, the config is uploaded to S3 in the EMMAA bucket.
            Default: False

        Returns
        -------
        dict
            A config data structure.
        """
        config = {
            # These are provided by the user upon initialization
            'name': self.name,
            'human_readable_name': self.human_readable_name,
            'description': self.description,
            # We don't make tests by default
            'make_tests': False,
            # We run daily upates by default
            'run_daily_update': True,
            # We first show the model just on dev
            'dev_only': True,
            # These are the search terms constructed upon
            # initialization
            'search_terms': [st.to_json()
                             for st in self.search_terms],
            # This is adopted from the template specified upon
            # initialization
            'assembly': self.assembly_config,
            # We configure the large corpus tests by default
            'test': {
                'statement_checking': {
                    'max_path_length': 10,
                    'max_paths': 1
                },
                'mc_types': [
                    'signed_graph', 'unsigned_graph'
                ],
                'make_links': True,
                'test_corpus': ['large_corpus_tests'],
                'default_test_corpus': 'large_corpus_tests',
                'filters': {
                    'large_corpus_tests': 'filter_chem_mesh_go'
                }
            }
        }
        if upload_to_s3:
            from emmaa.model import save_config_to_s3
            save_config_to_s3(self.name, config)
        return config

    def make_model(self, estmts, upload_to_s3=False):
        """Return, and optionally upload to S3 an initial EMMAA Model.

        Parameters
        ----------
        estmts : list of emmaa.statement.EmmaaStatement
            A list of prior EMMAA Statements to initialize the model with.
        upload_to_s3 : Optional[bool]
            If True, the model and the config are uploaded to S3, otherwise
            the model object is just returned without upload. Default: False

        Returns
        -------
        emmaa.model.EmmaaModel
            The EMMAA Model object constructed from the generated config
            and the given EMMAA Statements.
        """
        from emmaa.model import EmmaaModel
        config = self.make_config(upload_to_s3=upload_to_s3)
        model = EmmaaModel(name=self.name, config=config)
        model.add_statements(estmts)
        if upload_to_s3:
            model.save_to_s3()
        return model


def get_raw_statements_for_pmids(pmids, mode='all', batch_size=100):
    """Return EmmaaStatements based on extractions from given PMIDs.

    Parameters
    ----------
    pmids : set or list of str
        A set of PMIDs to find raw INDRA Statements for in the INDRA DB.
    mode : 'all' or 'distilled'
        The 'distilled' mode makes sure that the "best", non-redundant
        set of raw statements are found across potentially redundant text
        contents and reader versions. The 'all' mode doesn't do such
        distillation but is significantly faster.
    batch_size : Optional[int]
        Determines how many PMIDs to fetch statements for in each
        iteration. Default: 100.

    Returns
    -------
    dict
        A dict keyed by PMID with values INDRA Statements obtained
        from the given PMID.
    """
    db = get_db('primary')
    logger.info(f'Getting raw statements for {len(pmids)} PMIDs')
    all_stmts = defaultdict(list)
    for pmid_batch in tqdm.tqdm(batch_iter(pmids, return_func=set,
                                           batch_size=batch_size),
                                total=len(pmids)/batch_size):
        if mode == 'distilled':
            clauses = [
                db.TextRef.pmid.in_(pmid_batch),
                db.TextContent.text_ref_id == db.TextRef.id,
                db.Reading.text_content_id == db.TextContent.id,
                db.RawStatements.reading_id == db.Reading.id]
            distilled_stmts = distill_stmts(db, get_full_stmts=True,
                                            clauses=clauses)
            for stmt in distilled_stmts:
                all_stmts[stmt.evidence[0].pmid].append(stmt)
        else:
            id_stmts = \
                get_raw_stmt_jsons_from_papers(pmid_batch, id_type='pmid',
                                               db=db)
            for pmid, stmt_jsons in id_stmts.items():
                all_stmts[pmid] += stmts_from_json(stmt_jsons)
    all_stmts = dict(all_stmts)
    return all_stmts


def make_search_terms(search_strings, mesh_ids):
    """Return EMMAA SearchTerms based on search strings and MeSH IDs.

    Parameters
    ----------
    search_strings : list of str
        A list of search strings e.g., "diabetes" to find papers in the
        literature.
    mesh_ids : list of str
        A list of MeSH IDs that are used to search the literature as headings
        associated with papers.

    Returns
    -------
    list of emmmaa.prior.SearchTerm
        A list of EMMAA SearchTerm objects constructed from the search strings
        and the MeSH IDs.
    """
    search_terms = []
    for search_string in search_strings:
        search_term = SearchTerm(type='other', name=search_string,
                                 db_refs={}, search_term=search_string)
        search_terms.append(search_term)
    for mesh_id in mesh_ids:
        mesh_name = mesh_client.get_mesh_name(mesh_id)
        suffix = 'mh' if mesh_id.startswith('D') else 'nm'
        search_term = SearchTerm(type='mesh', name=mesh_name,
                                 db_refs={'MESH': mesh_id},
                                 search_term=f'{mesh_name} [{suffix}]')
        search_terms.append(search_term)
    return search_terms




"""The AWS Lambda emmaa-mm-update definition.

This file contains the function that updates model manager object in S3. It
must be placed on AWS Lambda, which can either be done manually (not
recommended) or by running:

$ python update_lambda.py model_manager_update.py emmaa-mm-update

in this directory.
"""

import boto3
from datetime import datetime

JOB_DEF = 'emmaa_jobdef'
QUEUE = 'emmaa-models-update-test'
PROJECT = 'aske'
PURPOSE = 'update-emmaa-model-manager'
BRANCH = 'origin/master'


def lambda_handler(event, context):
    """Create a batch job to update model manager on s3.

    This function is designed to be placed on AWS Lambda, taking the event and
    context arguments that are passed. Note that this function must always have
    the same parameters, even if any or all of them are unused, because we do
    not have control over what Lambda sends as parameters. This Lambda
    function is configured to be triggered when the model is updated on S3.

    See the top of the page for the Lambda update procedure.

    Parameters
    ----------
    event : dict
        A dictionary containing metadata regarding the triggering event. In
        this case, we are expecting 'Records', each of which contains a record
        of a file that was added (or changed) on s3.
    context : object
        This is an object containing potentially useful context provided by
        Lambda. See the documentation cited above for details.

    Returns
    -------
    ret : dict
        A dict containing 'statusCode', with a valid HTTP status code, 'result',
        and 'job_id' to be returned to Lambda.
    """
    batch = boto3.client('batch')
    records = event['Records']
    for rec in records:
        try:
            model_key = rec['s3']['object']['key']
        except KeyError:
            pass
        model_name = model_key.split('/')[1]
        core_command = 'bash scripts/git_and_run.sh'
        if BRANCH is not None:
            core_command += f' --branch {BRANCH}'
        core_command += (' python scripts/update_model_manager.py'
                         f' --model {model_name}')
        cont_overrides = {
            'command': ['python', '-m', 'indra.util.aws', 'run_in_batch',
                        '--project', PROJECT, '--purpose', PURPOSE,
                        core_command]
            }
        now_str = datetime.utcnow().strftime('%Y%m%d_%H%M%S')
        ret = batch.submit_job(jobName=f'{model_name}_mm_update_{now_str}',
                               jobQueue=QUEUE, jobDefinition=JOB_DEF,
                               containerOverrides=cont_overrides)
        job_id = ret['jobId']

    return {'statusCode': 200, 'result': 'SUCCESS', 'job_id': job_id}



import os
import re
import boto3
import logging
import json
import pickle
import zlib
import tweepy
from flask import Flask
from pathlib import Path
from datetime import datetime, timedelta
from botocore import UNSIGNED
from botocore.client import Config
from inflection import camelize
from zipfile import ZipFile
from indra.util.aws import get_s3_file_tree, get_date_from_str, iter_s3_keys
from indra.statements import get_all_descendants
from indra.literature.s3_client import gzip_string
from emmaa.subscription.email_service import email_bucket

FORMAT = '%Y-%m-%d-%H-%M-%S'
RE_DATETIMEFORMAT = r'\d{4}\-\d{2}\-\d{2}\-\d{2}\-\d{2}\-\d{2}'
RE_DATEFORMAT = r'\d{4}\-\d{2}\-\d{2}'
EMMAA_BUCKET_NAME = 'emmaa'
logger = logging.getLogger(__name__)


FORMATTED_TYPE_NAMES = {'pysb': 'PySB',
                        'pybel': 'PyBEL',
                        'signed_graph': 'Signed Graph',
                        'unsigned_graph': 'Unsigned Graph'}


def strip_out_date(keystring, date_format='datetime'):
    """Strips out datestring of selected date_format from a keystring"""
    if date_format == 'datetime':
        re_format = RE_DATETIMEFORMAT
    elif date_format == 'date':
        re_format = RE_DATEFORMAT
    try:
        return re.search(re_format, keystring).group()
    except AttributeError:
        logger.warning(f'Can\'t parse string {keystring} for date')
        return None


def make_date_str(date=None):
    """Return a date string in a standardized format.

    Parameters
    ----------
    date : Optional[datetime.datetime]
        A date object to get the standardized string for. If not provided,
        utcnow() is used to construct the date. (Note: using UTC is important
        because this code may run in multiple contexts).

    Returns
    -------
    str
        The datetime string in a standardized format.
    """
    if not date:
        date = datetime.utcnow()
    return date.strftime(FORMAT)


def list_s3_files(bucket, prefix, extension=None):
    client = get_s3_client()
    files = iter_s3_keys(client, bucket, prefix)
    if extension:
        keys = [f for f in files if f.endswith(extension)]
    else:
        keys = list(files)
    return keys


def sort_s3_files_by_date_str(bucket, prefix, extension=None):
    """
    Return the list of keys of the files on an S3 path sorted by date starting
    with the most recent one.
    """
    def process_key(key):
        fname_with_extension = os.path.basename(key)
        fname = os.path.splitext(fname_with_extension)[0]
        date_str = fname.split('_')[-1]
        return get_date_from_str(date_str)
    keys = list_s3_files(bucket, prefix, extension=extension)
    if len(keys) < 2:
        return keys
    keys = sorted(keys, key=lambda k: process_key(k), reverse=True)
    return keys


def sort_s3_files_by_last_mod(bucket, prefix, time_delta=None,
                              extension=None, unsigned=True, reverse=False,
                              w_dt=False):
    """Return a list of s3 object keys sorted by their LastModified date on S3

    Parameters
    ----------
    bucket : str
        s3 bucket to look for keys in
    prefix : str
        The prefix to use for the s3 keys
    time_delta : Optional[datetime.timedelta]
        If used, should specify how far back the to look for files on s3.
        Default: None
    extension : Optional[str]
        If used, limit keys to those with the matching file extension.
        Default: None.
    unsigned : bool
        If True, use unsigned s3 client. Default: True.
    reverse : bool
        Reverse the sort order of the returned s3 files. Default: False.
    w_dt : bool
        If True, return list with datetime object along with key as tuple
        (key, datetime.datetime). Default: False.

    Returns
    -------
    list
        A list of s3 keys. If w_dt is True, each item is a tuple of
        (key, datetime.datetime) of the LastModified date.
    """
    if time_delta is None:
        time_delta = timedelta()  # zero timedelta
    s3 = get_s3_client(unsigned)
    n_hours_ago = datetime.utcnow() - time_delta
    file_tree = get_s3_file_tree(s3, bucket, prefix,
                                 date_cutoff=n_hours_ago,
                                 with_dt=True)
    key_list = sorted(list(file_tree.get_leaves()), key=lambda t: t[1],
                      reverse=reverse)
    if extension:
        return [t if w_dt else t[0] for t in key_list
                if t[0].endswith(extension)]
    else:
        return key_list if w_dt else [t[0] for t in key_list]


def find_nth_latest_s3_file(n, bucket, prefix, extension=None):
    """Return the key of the file with nth (0-indexed) latest date string on
    an S3 path"""
    files = sort_s3_files_by_date_str(bucket, prefix, extension)
    try:
        latest = files[n]
        return latest
    except IndexError:
        logger.debug('File is not found.')


def find_latest_s3_file(bucket, prefix, extension=None):
    """Return the key of the file with latest date string on an S3 path"""
    return find_nth_latest_s3_file(0, bucket, prefix, extension)


def find_latest_s3_files(number_of_files, bucket, prefix, extension=None):
    """
    Return the keys of the specified number of files with latest date strings
    on an S3 path sorted by date starting with the earliest one.
    """
    files = sort_s3_files_by_date_str(bucket, prefix, extension)
    keys = []
    for ix in range(number_of_files):
        keys.append(files[ix])
    keys.reverse()
    return keys


def find_number_of_files_on_s3(bucket, prefix, extension=None):
    files = sort_s3_files_by_date_str(bucket, prefix, extension)
    return len(files)


def find_latest_emails(email_type, time_delta=None, w_dt=False):
    """Return a list of keys of the latest emails delivered to s3

    Parameters
    ----------
    email_type : str
        The email type to look for, e.g. 'feedback' if listing bounce and
        complaint emails sent to the ReturnPath address.
    time_delta : datetime.timedelta
        The timedelta to look backwards for listing emails.
    w_dt : bool
        If True, return a list of (key, datetime.datetime) tuples.

    Returns
    -------
    list[Keys]
        A list of keys to the emails of the specified type. If w_dt is True,
        each item is a tuple of (key, datetime.datetime) of the LastModified
        date.
    """
    email_list = sort_s3_files_by_last_mod(email_bucket, email_type,
                                           time_delta, unsigned=False,
                                           w_dt=w_dt)
    ignore = 'AMAZON_SES_SETUP_NOTIFICATION'
    if w_dt:
        return [s for s in email_list if ignore not in s[0]]
    return [s for s in email_list if ignore not in s]


def get_email_content(key):
    s3 = get_s3_client(unsigned=False)
    email_obj = s3.get_object(Bucket=email_bucket, Key=key)
    return email_obj['Body'].read().decode()


def find_index_of_s3_file(key, bucket, prefix, extension=None):
    files = sort_s3_files_by_date_str(bucket, prefix, extension)
    ix = files.index(key)
    return ix


def does_exist(bucket, prefix, extension=None):
    """Check if the file with exact key or starting with prefix and/or with
    extension exist in a bucket.
    """
    all_files = list_s3_files(bucket, prefix, extension)
    if any(fname.startswith(prefix) for fname in all_files):
        return True
    return False


def get_s3_client(unsigned=True):
    """Return a boto3 S3 client with optional unsigned config.

    Parameters
    ----------
    unsigned : Optional[bool]
        If True, the client will be using unsigned mode in which public
        resources can be accessed without credentials. Default: True

    Returns
    -------
    botocore.client.S3
        A client object to AWS S3.
    """
    if unsigned:
        return boto3.client('s3', config=Config(signature_version=UNSIGNED))
    else:
        return boto3.client('s3')


def get_class_from_name(cls_name, parent_cls):
    classes = get_all_descendants(parent_cls)
    for cl in classes:
        if cl.__name__.lower() == camelize(cls_name).lower():
            return cl
    raise NotAClassName(f'{cls_name} is not recognized as a '
                        f'{parent_cls.__name__} type!')


def _get_flask_app():
    emmaa_service_dir = Path(__file__).parent.parent.joinpath(
        'emmaa_service', 'templates')
    app = Flask('Static app', template_folder=emmaa_service_dir.as_posix())
    return app


def load_pickle_from_s3(bucket, key):
    client = get_s3_client()
    try:
        logger.info(f'Loading object from {key}')
        obj = client.get_object(Bucket=bucket, Key=key)
        content = pickle.loads(obj['Body'].read())
        return content
    except Exception as e:
        logger.info(f'Could not load the pickle from {key}')
        logger.info(e)


def save_pickle_to_s3(obj, bucket, key):
    client = get_s3_client(unsigned=False)
    logger.info('Pickling object')
    obj_str = pickle.dumps(obj, protocol=4)
    logger.info(f'Saving object to {key}')
    client.put_object(Body=obj_str, Bucket=bucket, Key=key)


def load_json_from_s3(bucket, key):
    client = get_s3_client()
    logger.info(f'Loading object from {key}')
    obj = client.get_object(Bucket=bucket, Key=key)
    content = json.loads(obj['Body'].read().decode('utf8'))
    return content


def save_json_to_s3(obj, bucket, key, save_format='json'):
    client = get_s3_client(unsigned=False)
    json_str = _get_json_str(obj, save_format=save_format)
    logger.info(f'Uploading the {save_format} object to S3')
    client.put_object(Body=json_str.encode('utf8'),
                      Bucket=bucket, Key=key)


def load_gzip_json_from_s3(bucket, key):
    client = get_s3_client()
    # Newer files are zipped with gzip while older with zipfile
    try:
        logger.info(f'Loading zipped object from {key}')
        gz_obj = client.get_object(Bucket=bucket, Key=key)
        content = json.loads(zlib.decompress(
            gz_obj['Body'].read(), 16+zlib.MAX_WBITS).decode('utf8'))
    except Exception as e:
        logger.info(e)
        logger.info('Could not load with gzip, using zipfile')
        logger.info(f'Loading zipfile from {key}')
        client.download_file(bucket, key, 'temp.zip')
        with ZipFile('temp.zip', 'r') as zipf:
            content = json.loads(zipf.read(zipf.namelist()[0]))
    return content


def save_gzip_json_to_s3(obj, bucket, key, save_format='json'):
    client = get_s3_client(unsigned=False)
    json_str = _get_json_str(obj, save_format=save_format)
    gz_str = gzip_string(json_str, f'assembled_stmts.{save_format}')
    client.put_object(Body=gz_str, Bucket=bucket, Key=key)


def _get_json_str(json_obj, save_format='json'):
    logger.info(f'Dumping the {save_format} into a string')
    if save_format == 'json':
        json_str = json.dumps(json_obj, indent=1)
    elif save_format == 'jsonl':
        json_str = '\n'.join(
            [json.dumps(item) for item in json_obj])
    return json_str


class EmailHtmlBody(object):
    app = _get_flask_app()

    def __init__(self, domain='emmaa.indra.bio',
                 template_path='email_unsub/email_body.html'):
        self.template = self.app.jinja_env.get_template(template_path)
        self.domain = domain
        self.static_tab_link = f'https://{domain}/query?tab=static'
        self.dynamic_tab_link = f'https://{domain}/query?tab=dynamic'
        self.open_tab_link = f'https://{domain}/query?tab=open'

    def render(self, static_query_deltas, open_query_deltas,
               dynamic_query_deltas, unsub_link):
        """Provided the delta json objects, render HTML to put in email body

        Parameters
        ----------
        static_query_deltas : json
            A list of lists that names which queries have updates. Expected
            structure:
            [(english_query, detailed_query_link, model, model_type)]
        dynamic_query_deltas : list[
            A list of lists that names which queries have updates. Expected
            structure:
            [(english_query, model, model_type)]
        unsub_link : str

        Returns
        -------
        html
            An html string rendered from the associated jinja2 template
        """
        if not static_query_deltas and not open_query_deltas and \
                not dynamic_query_deltas:
            raise ValueError('No query deltas provided')
        # Todo consider generating unsubscribe link here, will probably have
        #  to solve import loops for that though
        return self.template.render(
            static_tab_link=self.static_tab_link,
            static_query_deltas=static_query_deltas,
            open_tab_link=self.open_tab_link,
            open_query_deltas=open_query_deltas,
            dynamic_tab_link=self.dynamic_tab_link,
            dynamic_query_deltas=dynamic_query_deltas,
            unsub_link=unsub_link
        ).replace('\n', '')


class NotAClassName(Exception):
    pass


def get_credentials(key):
    client = boto3.client('ssm')
    auth_dict = {}
    for par in ['consumer_token', 'consumer_secret', 'access_token',
                'access_secret']:
        name = f'/twitter/{key}/{par}'
        try:
            response = client.get_parameter(Name=name, WithDecryption=True)
            val = response['Parameter']['Value']
            auth_dict[par] = val
        except Exception as e:
            print(e)
            break
    return auth_dict


def get_oauth_dict(auth_dict):
    oauth = tweepy.OAuthHandler(auth_dict.get('consumer_token'),
                                auth_dict.get('consumer_secret'))
    oauth.set_access_token(auth_dict.get('access_token'),
                           auth_dict.get('access_secret'))
    return oauth


def update_status(msg, twitter_cred):
    twitter_auth = get_oauth_dict(twitter_cred)
    if twitter_auth is None:
        return
    twitter_api = tweepy.API(twitter_auth)
    twitter_api.update_status(msg)



import logging
from datetime import datetime
from copy import deepcopy

from emmaa.model_tests import load_model_manager_from_s3
from emmaa.db import get_db
from emmaa.util import make_date_str, find_latest_s3_file, EmailHtmlBody, \
    EMMAA_BUCKET_NAME, FORMATTED_TYPE_NAMES
from emmaa.subscription.email_util import generate_unsubscribe_link


logger = logging.getLogger(__name__)


model_manager_cache = {}


email_html = EmailHtmlBody()


class QueryManager(object):
    """Manager to run queries and interact with the database.

    Parameters
    ----------
    db : emmaa.db.EmmaaDatabaseManager
        An instance of a database manager to use.
    model_managers : list[emmaa.model_tests.ModelManager]
        Optional list of ModelManagers to use for running queries. If not
        given, the methods will load ModelManager from S3 when needed.
    """
    def __init__(self, db=None, model_managers=None):
        self.db = db
        if db is None:
            self.db = get_db('primary')
        self.model_managers = model_managers if model_managers else []

    def answer_immediate_query(
            self, user_email, user_id, query, model_names, subscribe,
            use_kappa=False, bucket=EMMAA_BUCKET_NAME):
        """This method first tries to find saved result to the query in the
        database and if not found, runs ModelManager method to answer query."""
        query_type = query.get_type()
        # Retrieve query-model hashes
        query_hashes = [
            query.get_hash_with_model(model) for model in model_names]
        # Store query in the database for future reference.
        self.db.put_queries(user_email, user_id, query, model_names, subscribe)
        # Check if the query has already been answered for any of given models
        # and retrieve the results from database.
        saved_results = self.db.get_results_from_query(query, model_names)
        if not saved_results:
            saved_results = []
        checked_models = {res[0] for res in saved_results}
        # If the query was answered for all models before, return the hashes.
        if checked_models == set(model_names):
            return {query_type: query_hashes}
        # Run queries mechanism for models for which result was not found.
        new_results = []
        new_date = datetime.now()
        for model_name in model_names:
            if model_name not in checked_models:
                results_to_store = []
                mm = self.get_model_manager(model_name)
                response_list = mm.answer_query(
                    query, use_kappa=use_kappa, bucket=bucket)
                for (mc_type, response) in response_list:
                    results_to_store.append((query, mc_type, response))
                self.db.put_results(model_name, results_to_store)
        return {query_type: query_hashes}

    def answer_registered_queries(self, model_name, use_kappa=False,
                                  bucket=EMMAA_BUCKET_NAME):
        """Retrieve and asnwer registered queries

        Retrieve queries registered on database for a given model,
        answer them, calculate delta between results and put results to a
        database.

        Parameters
        ----------
        model_name : str
            The name of the model
        use_kappa : bool
            If True, use kappa modeling when answering the dynamic query
        bucket : str
            The bucket to save the results to
        """
        model_manager = self.get_model_manager(model_name)
        queries = self.db.get_queries(model_name)
        logger.info(f'Found {len(queries)} queries for {model_name} model.')
        # Only do the following steps if there are queries for this model
        if queries:
            results = model_manager.answer_queries(
                queries, use_kappa=use_kappa, bucket=bucket)
            new_results = [(model_name, result[0], result[1], result[2], '')
                           for result in results]
            self.db.put_results(model_name, results)

    def get_registered_queries(self, user_email, query_type='path_property'):
        """Get formatted results to queries registered by user."""
        results = self.db.get_results(user_email, query_type=query_type)
        return format_results(results, query_type)

    def retrieve_results_from_hashes(
            self, query_hashes, query_type='path_property', latest_order=1):
        """Retrieve results from a db given a list of query-model hashes."""
        results = self.db.get_results_from_hashes(
            query_hashes, latest_order=latest_order)
        return format_results(results, query_type)

    def get_user_query_delta(self, user_email, domain='emmaa.indra.bio'):
        """Produce a report for all query results per user in a given format

        Parameters
        ----------
        user_email : str
            The email of the user for which to get the report for
        domain : str
            The domain name for the unsubscibe link in the html
            report. Default: "emmaa.indra.bio".

        Returns
        -------
        tuple(str, html_str)
            A tuple with (str report, html report)
        """
        logger.info(f'Finding query delta for {user_email}')
        # Get results of user's query
        results = self.db.get_results(user_email, latest_order=1)

        # Get the query deltas
        static_results_delta, open_results_delta, dynamic_results_delta = \
            make_reports_from_results(results, domain=domain)
        # Make text report
        str_report = make_str_report_per_user(static_results_delta,
                                              open_results_delta,
                                              dynamic_results_delta)
        str_report = str_report if str_report else ''

        # Make html report
        html_report = make_html_report_per_user(static_results_delta,
                                                open_results_delta,
                                                dynamic_results_delta,
                                                user_email,
                                                domain=domain)
        html_report = html_report if html_report else None

        if html_report:
            logger.info(f'Found query delta for {user_email}')
        else:
            logger.info(f'No query delta to report for {user_email}')
        return str_report, html_report

    def get_model_manager(self, model_name):
        # Try get model manager from class attributes or load from s3.
        for mm in self.model_managers:
            if mm.model.name == model_name:
                return mm
        return load_model_manager_from_cache(model_name)


def make_reports_from_results(new_results, domain='emmaa.indra.bio'):
    """Make a report given latest results and queries the results are for.

    Parameters
    ----------
    new_results : list[tuple]
        Latest results as a list of tuples where each tuple has the format
        (model_name, query, mc_type, result_json, date, delta).

    Returns
    -------
    reports : list
        A list of reports on changes for each of the queries.
    """
    processed_query_mc = []
    static_reports = []
    open_reports = []
    dynamic_reports = []
    for model_name, query, mc_type, result_json, delta, _ in new_results:
        if (model_name, query, mc_type) in processed_query_mc:
            continue
        if delta:
            model_type_name = FORMATTED_TYPE_NAMES[
                mc_type] if mc_type else mc_type
            rep = [
                query.to_english(),
                _detailed_page_link(
                    domain,
                    model_name,
                    mc_type,
                    query.get_hash_with_model(
                        model_name)),
                model_name,
                model_type_name
            ]
            # static
            if query.get_type() == 'path_property':
                static_reports.append(rep)
            # open
            elif query.get_type() == 'open_search_query':
                open_reports.append(rep)
            # dynamic
            else:
                # Remove link for dynamic
                _ = rep.pop(1)
                dynamic_reports.append(rep)
        processed_query_mc.append((model_name, query, mc_type))
    return static_reports, open_reports, dynamic_reports


def make_str_report_per_user(static_results_delta, open_results_delta,
                             dynamic_results_delta):
    """Produce a report for all query results per user as a string.

    Parameters
    ----------
    static_results_delta : list
        A list of tuples of query deltas for static queries. Each tuple
        has a format (english_query, link, model, mc_type)
    open_results_delta : list
        A list of tuples of query deltas for open queries. Each tuple
        has a format (english_query, link, model, mc_type)
    dynamic_results_delta : list
        A list of tuples of query deltas for dynamic queries. Each tuple
        has a format (english_query, link, model, mc_type) (no link in
        dynamic_results_delta tuples).

    Returns
    -------
    msg : str
        A message about query deltas.
    """
    if not static_results_delta and not open_results_delta and not \
            dynamic_results_delta:
        logger.info('No delta provided')
        return None
    msg = ''
    if static_results_delta:
        msg += 'Updates to your static queries:\n'
        for english_query, _, model, mc_type in static_results_delta:
            msg += f'{english_query} in {model} using the {mc_type}.\n'
    if open_results_delta:
        msg += 'Updates to your open queries:\n'
        for english_query, _, model, mc_type in open_results_delta:
            msg += f'{english_query} in {model} using the {mc_type}.\n'
    if dynamic_results_delta:
        msg += 'Updates to your dynamic queries:\n'
        for english_query, model, mc_type in dynamic_results_delta:
            msg += f'{english_query} in {model} using the {mc_type}.\n'
    return msg


def make_html_report_per_user(static_results_delta, open_results_delta,
                              dynamic_results_delta, email,
                              domain='emmaa.indra.bio'):
    """Produce a report for all query results per user in an html file.

    Parameters
    ----------
    static_results_delta : list
        A list of tuples of query deltas for static queries. Each tuple
        has a format (english_query, link, model, mc_type)
    open_results_delta : list
        A list of tuples of query deltas for open queries. Each tuple
        has a format (english_query, link, model, mc_type)
    dynamic_results_delta : list
        A list of tuples of query deltas for dynamic queries. Each tuple
        has a format (english_query, link, model, mc_type)
    email : str
        The email of the user to get the results for.
    domain : str
        The domain name for the unsubscibe link in the report. Default:
        "emmaa.indra.bio".

    Returns
    -------
    str
        A string containing an html document
    """
    # Generate unsubscribe link
    link = generate_unsubscribe_link(email=email, domain=domain)

    if static_results_delta or open_results_delta or dynamic_results_delta:
        return email_html.render(
            static_query_deltas=static_results_delta,
            open_query_deltas=open_results_delta,
            dynamic_query_deltas=dynamic_results_delta,
            unsub_link=link
        )
    else:
        return ''


def _detailed_page_link(domain, model_name, model_type, query_hash):
    # example:
    # https://emmaa.indra.bio/query/aml/?model_type=pysb&query_hash
    # =4911955502409811&order=1
    return f'https://{domain}/query/{model_name}?model_type=' \
           f'{model_type}&query_hash={query_hash}&order=1'


def format_results(results, query_type='path_property'):
    """Format db output to a standard json structure."""
    model_types = ['pysb', 'pybel', 'signed_graph', 'unsigned_graph']
    formatted_results = {}
    for result in results:
        model = result[0]
        query = result[1]
        query_hash = query.get_hash_with_model(model)
        if query_hash not in formatted_results:
            formatted_results[query_hash] = {
                'query': query.to_english(),
                'model': model,
                'date': make_date_str(result[5])}
        mc_type = result[2]
        response_json = result[3]
        delta = result[4]
        response = []
        for k, v in response_json.items():
            if isinstance(v, str):
                response = v
            elif isinstance(v, dict):
                if k in delta:
                    new_v = deepcopy(v)
                    new_v['path'] = ('new', new_v['path'])
                    response.append(new_v)
                else:
                    response.append(v)
        if query_type in ['path_property', 'open_search_query']:
            if mc_type == '' and \
                    response == 'Query is not applicable for this model':
                for mt in model_types:
                    formatted_results[query_hash][mt] = ['n_a', response]
            elif isinstance(response, str) and \
                    response == 'Statement type not handled':
                formatted_results[query_hash][mc_type] = ['n_a', response]
            elif isinstance(response, str) and \
                    not response == 'Path found but exceeds search depth':
                formatted_results[query_hash][mc_type] = ['Fail', response]
            else:
                formatted_results[query_hash][mc_type] = ['Pass', response]
        elif query_type == 'dynamic_property':
            if response == 'Query is not applicable for this model':
                formatted_results[query_hash]['result'] = ['n_a', response]
            else:
                res = int(response[0]['sat_rate'] * 100)
                expl = (f'Satisfaction rate is {res}% after '
                        f'{response[0]["num_sim"]} simulations.')
                if res > 50:
                    formatted_results[query_hash]['result'] = ['Pass', expl]
                else:
                    formatted_results[query_hash]['result'] = ['Fail', expl]
                formatted_results[query_hash]['image'] = (
                    response[0]['fig_path'])
    if query_type in ['path_property', 'open_search_query']:
        # Loop through the results again to make sure all model types are there
        for qh in formatted_results:
            for mt in model_types:
                if mt not in formatted_results[qh]:
                    formatted_results[qh][mt] = [
                        'n_a', 'Model type not supported']
    return formatted_results


def load_model_manager_from_cache(model_name, bucket=EMMAA_BUCKET_NAME):
    model_manager = model_manager_cache.get(model_name)
    if model_manager:
        latest_on_s3 = find_latest_s3_file(
            bucket, f'results/{model_name}/model_manager_', '.pkl')
        cached_date = model_manager.date_str
        logger.info(f'Found model manager cached on {cached_date} and '
                    f'latest file on S3 is {latest_on_s3}')
        if cached_date in latest_on_s3:
            logger.info(f'Loaded model manager for {model_name} from cache.')
            return model_manager
    logger.info(f'Loading model manager for {model_name} from S3.')
    model_manager = load_model_manager_from_s3(
        model_name=model_name, bucket=bucket)
    model_manager_cache[model_name] = model_manager
    return model_manager


def answer_queries_from_s3(model_name, db=None, bucket=EMMAA_BUCKET_NAME):
    """Answer registered queries with model manager on s3.

    Parameters
    ----------
    model_name : str
        Name of EmmaaModel to answer queries for.
    db : Optional[emmaa.db.manager.EmmaaDatabaseManager]
        If given over-rides the default primary database.
    """
    mm = load_model_manager_from_s3(model_name=model_name, bucket=bucket)
    qm = QueryManager(db=db, model_managers=[mm])
    qm.answer_registered_queries(model_name)



import os
import hmac
import hashlib
import logging
from urllib import parse
from datetime import datetime, timedelta

from emmaa.db import get_db

db = get_db('primary')

logger = logging.getLogger(__name__)

EMAIL_SIGNATURE_KEY = os.environ.get('EMAIL_SIGN_SECRET')


def __sign_str_concat(email, expiration_str):
    """This is the method to concatenate strings that are to be used in HMAC
    signature generation.

    Email should NOT be url encoded.
    """
    return ' '.join([email, expiration_str])


def generate_unsubscribe_qs(email, days=7):
    """Generate an unsubscribe query string for a url

    Parameters
    ----------
    email : str
        A valid email address
    days : int
        The number of days the query string should be valid. Default: 7.

    Returns
    -------
    str
        A query string of the format 'email=<urlenc
        email>&expiration=<timestamp>&signature=<sha256 hex>'
    """
    if days < 1:
        logger.warning('Expiration date is less than one day into the '
                       'future. Link will likely already be expired.')
    future = datetime.utcnow() + timedelta(days=days)
    expiration = str(future.timestamp()).split('.')[0]
    signature = generate_signature(email=email, expire_str=expiration)
    return parse.urlencode({'email': email,
                            'expiration': expiration,
                            'signature': signature})


def generate_unsubscribe_link(email, days=7, domain='emmaa.indra.bio'):
    """Generate an unsubscribe link for the provided email address

    Given an email address, generate an unsubscribe link using that email
    address. Optionally provide the number of days into the future the link
    should be valid until and the domain name. The domain name is expeceted
    to be of the format "some.domain.com". The appropriate path and prefixes
    will be added together with the query string. Example:

    >>> generate_unsubscribe_link('user@email.com', domain='some.domain.com')
    >>> 'https://some.domain.com/query/unsubscribe?email=user%40email.com' +
        '&expiration=1234567890&signature=1234567890abcdef'

    Parameters
    ----------
    email : str
        An email address.
    days : int
        The number of days into the future the link should be valid until.
        Default: 7.
    domain : str
        A domain name to prefix the query string with. Expected format is:
        "some.domain.com". Default: 'emmaa.indra.bio'

    Returns
    -------
    str
        An unsubscribe link for the provided email and (optionally) domain
    """
    qs = generate_unsubscribe_qs(email, days)
    link = f'https://{domain}/query/unsubscribe?{qs}'
    return link


def generate_signature(email, expire_str, digestmod=hashlib.sha256):
    """Return an HMAC signature based on email and expire_str

    From documentation of HMAC in python:
    key is a bytes or bytearray object giving the secret key.
    If msg is present, the method call update(msg) is made.
    digestmod is the digest name, digest constructor or module for the HMAC
    object to use. It supports any name suitable to hashlib.new().

    Parameters
    ----------
    email : str
        A valid email address. Should not be URL encoded.
    expire_str : str
        A timestamp string in seconds
    digestmod : str|digest constructor|module
        digest name, digest constructor or module for the HMAC object to
        use. Default: hashlib.sha256

    Returns
    -------
    str
        A hexadecimal string representing the signature
    """
    if not EMAIL_SIGNATURE_KEY:
        raise ValueError('No secret key set for email signature. '
                         'Cannot generate signature')

    digester = hmac.new(key=EMAIL_SIGNATURE_KEY.encode(encoding='utf-8'),
                        msg=__sign_str_concat(
                            email, expire_str).encode(encoding='utf-8'),
                        digestmod=digestmod)
    return digester.hexdigest()


def verify_email_signature(signature, email, expiration, 
                           digestmod=hashlib.sha256):
    """Verify HMAC signature"""
    if not EMAIL_SIGNATURE_KEY:
        logger.error('No secret key set for email signature. '
                     'Cannot verify signature')
        return False
    actual_digest = hmac.new(
        key=EMAIL_SIGNATURE_KEY.encode(encoding='utf-8'),
        msg=__sign_str_concat(email, expiration).encode(encoding='utf-8'),
        digestmod=digestmod).hexdigest()

    if len(signature) != len(actual_digest):
        return False
    try:
        return hmac.compare_digest(actual_digest, signature)
    except Exception:
        return False


def get_email_subscriptions(email):
    """Verifies which email subsciptions exist for the provided email

    Parameters
    ----------
    email : str
        The email to the check subscriptions for

    Returns
    -------
    list(tuple(str, str, query_hash))
    """
    user_queries = db.get_subscribed_queries(email)
    if not user_queries:
        return []
    return [(qo.to_english() + f' for model {mid}',
             f'{qo.get_type()}'.replace('_', ' '), qh)
            for qo, mid, qh in user_queries]


def register_email_unsubscribe(email, queries):
    """Executes an email unsubscribe request"""
    success = db.update_email_subscription(email, queries, False)
    return success



import logging
import jsonpickle
from collections import defaultdict
from emmaa.model import load_config_from_s3, get_model_stats, \
    load_stmts_from_s3
from emmaa.model_tests import load_model_manager_from_s3
from emmaa.util import find_latest_s3_file, find_nth_latest_s3_file, \
    strip_out_date, EMMAA_BUCKET_NAME, load_json_from_s3, save_json_to_s3, \
    get_credentials, update_status
from indra.statements.statements import Statement
from indra.assemblers.english.assembler import EnglishAssembler
from indra.literature import pubmed_client, crossref_client, pmc_client
from indra_db import get_db
from indra_db.client.principal.curation import get_curations
from indra_db.util import unpack


logger = logging.getLogger(__name__)


CONTENT_TYPE_FUNCTION_MAPPING = {
    'statements': 'get_stmt_hashes',
    'applied_tests': 'get_applied_test_hashes',
    'passed_tests': 'get_passed_test_hashes',
    'paths': 'get_passed_test_hashes',
    'raw_papers': 'get_all_raw_paper_ids',
    'assembled_papers': 'get_all_assembled_paper_ids'}


TWITTER_MODEL_TYPES = {'pysb': '@PySysBio',
                       'pybel': '@pybelbio',
                       'signed_graph': 'Signed Graph',
                       'unsigned_graph': 'Unsigned Graph'}


class Round(object):
    """Parent class for classes analyzing one round of something (model or
    tests).

    Parameters
    ----------
    date_str : str
        Time when ModelManager responsible for this round was created.

    Attributes
    ----------

    function_mapping : dict
        A dictionary of strings mapping a type of content to a tuple of
        functions necessary to find delta for this type of content. First
        function in a tuple gets a list of all hashes for a given content type,
        while the second returns an English description of a given content type
        for a single hash.
    """
    def __init__(self, date_str):
        self.date_str = date_str
        self.function_mapping = CONTENT_TYPE_FUNCTION_MAPPING

    @classmethod
    def load_from_s3_key(cls, key):
        raise NotImplementedError("Method must be implemented in child class.")        

    def get_english_statement(self, stmt):
        ea = EnglishAssembler([stmt])
        sentence = ea.make_model()
        return ('', sentence, '')

    def find_delta_hashes(self, other_round, content_type, **kwargs):
        """Return a dictionary of changed hashes of a given content type. This
        method makes use of self.function_mapping dictionary.

        Parameters
        ----------
        other_round : emmaa.analyze_tests_results.TestRound
            A different instance of a TestRound
        content_type : str
            A type of the content to find delta. Accepted values:
            - statements
            - applied_tests
            - passed_tests
            - paths
        **kwargs : dict
            For some of content types, additional arguments must be
            provided sych as mc_type.
        Returns
        -------
        hashes : dict
            A dictionary containing lists of added and removed hashes of a
            given content type between two test rounds.
        """
        logger.info(f'Finding a hashes delta for {content_type}.')
        latest_hashes = getattr(
            self, self.function_mapping[content_type])(**kwargs)
        logger.info(f'Found {len(latest_hashes)} hashes in current round.')
        previous_hashes = getattr(
            other_round,
            other_round.function_mapping[content_type])(**kwargs)
        logger.info(f'Found {len(previous_hashes)} hashes in other round.')
        # Find hashes unique for each of the rounds - this is delta
        added_hashes = list(set(latest_hashes) - set(previous_hashes))
        removed_hashes = list(set(previous_hashes) - set(latest_hashes))
        hashes = {'added': added_hashes, 'removed': removed_hashes}
        return hashes


class ModelRound(Round):
    """Analyzes the results of one model update round.

    Parameters
    ----------
    statements : list[indra.statements.Statement]
        A list of INDRA Statements used to assemble a model.
    date_str : str
        Time when ModelManager responsible for this round was created.
    paper_ids : list(str)
        A list of paper IDs used to get raw statements for this round.
    paper_id_type : str
        Type of paper ID used.

    Attributes
    ----------
    stmts_by_papers : dict
        A dictionary mapping the paper IDs to sets of hashes of assembled
        statements with evidences retrieved from these papers.
    """
    def __init__(self, statements, date_str, paper_ids=None,
                 paper_id_type='TRID', emmaa_statements=None):
        super().__init__(date_str)
        self.statements = statements
        self.paper_ids = paper_ids if paper_ids else []
        self.paper_id_type = paper_id_type
        self.emmaa_statements = emmaa_statements if emmaa_statements else []
        self.stmts_by_papers = self.get_assembled_stmts_by_paper(paper_id_type)

    @classmethod
    def load_from_s3_key(cls, key, bucket=EMMAA_BUCKET_NAME,
                         load_estmts=False):
        mm = load_model_manager_from_s3(key=key, bucket=bucket)
        if not mm:
            return
        statements = mm.model.assembled_stmts
        date_str = mm.date_str
        try:
            paper_ids = list(mm.model.paper_ids)
        except AttributeError:
            paper_ids = None
        paper_id_type = mm.model.reading_config.get('main_id_type', 'TRID')
        estmts = None
        if load_estmts:
            estmts, _ = load_stmts_from_s3(mm.model.name, bucket)
        return cls(statements, date_str, paper_ids, paper_id_type, estmts)

    def get_total_statements(self):
        """Return a total number of statements in a model."""
        total = len(self.statements)
        logger.info(f'An assembled model has {total} statements.')
        return total

    def get_stmt_hashes(self):
        """Return a list of hashes for all statements in a model."""
        return [str(stmt.get_hash(refresh=True)) for stmt in self.statements]

    def get_statement_types(self):
        """Return a sorted list of tuples containing a statement type and a
        number of times a statement of this type occured in a model.
        """
        statement_types = defaultdict(int)
        logger.info('Finding a distribution of statements types.')
        for stmt in self.statements:
            statement_types[type(stmt).__name__] += 1
        return sorted(statement_types.items(), key=lambda x: x[1], reverse=True)

    def get_agent_distribution(self):
        """Return a sorted list of tuples containing an agent name and a number
        of times this agent occured in statements of a model."""
        logger.info('Finding agent distribution among model statements.')
        agent_count = defaultdict(int)
        for stmt in self.statements:
            for agent in stmt.agent_list():
                if agent is not None:
                    agent_count[agent.name] += 1
        return sorted(agent_count.items(), key=lambda x: x[1], reverse=True)

    def get_statements_by_evidence(self):
        """Return a sorted list of tuples containing a statement hash and a
        number of times this statement occured in a model."""
        stmts_evidence = {}
        for stmt in self.statements:
            stmts_evidence[str(stmt.get_hash(refresh=True))] = len(stmt.evidence)
        logger.info('Sorting statements by evidence count.')
        return sorted(stmts_evidence.items(), key=lambda x: x[1], reverse=True)

    def get_english_statements_by_hash(self):
        """Return a dictionary mapping a statement and its English description."""
        stmts_by_hash = {}
        for stmt in self.statements:
            stmts_by_hash[str(stmt.get_hash(refresh=True))] = (
                self.get_english_statement(stmt))
        return stmts_by_hash

    def get_sources_distribution(self):
        logger.info('Finding distribution of sources of statement evidences.')
        sources_count = defaultdict(int)
        for stmt in self.statements:
            for evid in stmt.evidence:
                if evid.source_api:
                    sources_count[evid.source_api] += 1
        return sorted(sources_count.items(), key=lambda x: x[1], reverse=True)

    def get_all_raw_paper_ids(self):
        """Return all paper IDs used in this round."""
        return self.paper_ids

    def get_number_raw_papers(self):
        """Return a total number of papers in this round."""
        return len(self.paper_ids)

    def get_assembled_stmts_by_paper(self, id_type='TRID'):
        """Get a mapping of paper IDs (TRID or PII) to assembled statements."""
        logger.info('Mapping papers to statements')
        stmts_by_papers = {}
        for stmt in self.statements:
            stmt_hash = stmt.get_hash()
            for evid in stmt.evidence:
                paper_id = None
                if id_type == 'pii':
                    paper_id = evid.annotations.get('pii')
                if evid.text_refs:
                    paper_id = evid.text_refs.get(id_type)
                    if not paper_id:
                        paper_id = evid.text_refs.get(id_type.lower())
                if paper_id:
                    if paper_id in stmts_by_papers:
                        stmts_by_papers[paper_id].add(stmt_hash)
                    else:
                        stmts_by_papers[paper_id] = {stmt_hash}
        for k, v in stmts_by_papers.items():
            stmts_by_papers[k] = list(v)
        return stmts_by_papers

    def get_all_assembled_paper_ids(self):
        return list(self.stmts_by_papers.keys())

    def get_number_assembled_papers(self):
        return len(self.stmts_by_papers)

    def get_papers_distribution(self):
        """Return a sorted list of tuples containing a paper ID and a number
        of unique statements extracted from that paper."""
        logger.info('Finding paper distribution')
        paper_stmt_count = {paper_id: len(stmts) for (paper_id, stmts) in
                            self.stmts_by_papers.items()}
        return sorted(paper_stmt_count.items(), key=lambda x: x[1],
                      reverse=True)

    def get_raw_paper_counts(self):
        logger.info('Finding raw statement count per paper')
        if not self.emmaa_statements:
            logger.info('Did not load raw EMMAA statements')
            return {}
        raw_by_papers = defaultdict(int)
        for estmt in self.emmaa_statements:
            for evid in estmt.stmt.evidence:
                paper_id = None
                id_type = self.paper_id_type
                if id_type == 'pii':
                    paper_id = evid.annotations.get('pii')
                if evid.text_refs:
                    paper_id = evid.text_refs.get(id_type)
                    if not paper_id:
                        paper_id = evid.text_refs.get(id_type.lower())
                if paper_id:
                    raw_by_papers[paper_id] += 1
        return raw_by_papers

    def get_paper_titles_and_links(self, trids):
        """Return a dictionary mapping paper IDs to their titles."""
        if self.paper_id_type == 'pii':
            return {}, {}
        db = get_db('primary')
        trs = db.select_all(db.TextRef, db.TextRef.id.in_(trids))
        ref_dicts = [tr.get_ref_dict() for tr in trs]
        trid_to_title = {}
        trid_to_link = {}
        trid_to_pmids = {}
        trid_to_pmcids = {}
        trid_to_dois = {}
        check_in_db = []
        # Map TRIDs to available PMIDs, DOIs, PMCIDs in this order
        for ref_dict in ref_dicts:
            if 'PMID' in ref_dict:
                trid_to_pmids[ref_dict['TRID']] = ref_dict['PMID']
            elif 'PMCID' in ref_dict:
                trid_to_pmcids[ref_dict['TRID']] = ref_dict['PMCID']
            elif 'DOI' in ref_dict:
                trid_to_dois[ref_dict['TRID']] = ref_dict['DOI']

        logger.info(f'From {len(trids)} TRIDs got {len(trid_to_pmids)} PMIDs,'
                    f' {len(trid_to_pmcids)} PMCIDs, {len(trid_to_dois)} DOIs')

        # First get titles and links for available PMIDs
        if trid_to_pmids:
            logger.info(f'Getting titles for {len(trid_to_pmids)} PMIDs')
            pmids = list(trid_to_pmids.values())
            pmids_to_titles = _get_pmid_titles(pmids)

            for trid, pmid in trid_to_pmids.items():
                if pmid in pmids_to_titles:
                    trid_to_title[str(trid)] = pmids_to_titles[pmid]
                else:
                    check_in_db.append(trid)
                link = _get_publication_link(pmid, 'PMID')
                trid_to_link[str(trid)] = link

        # Then get titles and links for available PMCIDs
        if trid_to_pmcids:
            logger.info(f'Getting titles for {len(trid_to_pmcids)} PMCIDs')
            for trid, pmcid in trid_to_pmcids.items():
                title = _get_pmcid_title(pmcid)
                if title:
                    trid_to_title[str(trid)] = title
                else:
                    check_in_db.append(trid)
                link = _get_publication_link(pmcid, 'PMCID')
                trid_to_link[str(trid)] = link

        # Then get titles and links for available DOIs
        if trid_to_dois:
            logger.info(f'Getting titles for {len(trid_to_dois)} DOIs')
            for trid, doi in trid_to_dois.items():
                title = _get_doi_title(doi)
                if title:
                    trid_to_title[str(trid)] = title
                else:
                    check_in_db.append(trid)
                link = _get_publication_link(doi, 'DOI')
                trid_to_link[str(trid)] = link

        # Try getting remaining titles from db
        if check_in_db:
            logger.info(f'Getting titles for {len(check_in_db)} remaining '
                        'TRIDs from DB')
            tcs = db.select_all(db.TextContent,
                                db.TextContent.text_ref_id.in_(check_in_db),
                                db.TextContent.text_type == 'title')
            for tc in tcs:
                title = unpack(tc.content)
                trid_to_title[str(tc.text_ref_id)] = title
        return trid_to_title, trid_to_link

    def get_curation_stats(self):
        if not self.emmaa_statements:
            logger.info('Did not load raw EMMAA statements')
            return
        curations = get_curations()
        curators_ev = defaultdict(set)
        curators_stmt = defaultdict(set)
        curators_ev_counts = {}
        curators_stmt_counts = {}
        curs_by_tags = defaultdict(int)
        curs_by_hash = defaultdict(list)
        cur_ev_dates = defaultdict(set)
        cur_stmt_dates = defaultdict(set)
        cur_ev_date_sum = []
        cur_stmt_date_sum = []
        for cur in curations:
            curs_by_hash[cur['source_hash']].append(cur)
        df = '%Y-%m-%d-00-00-00'
        for estmt in self.emmaa_statements:
            for ev in estmt.stmt.evidence:
                source_hash = ev.get_source_hash()
                curs_for_hash = curs_by_hash.get(source_hash)
                if curs_for_hash:
                    for cur in curs_for_hash:
                        curators_ev[cur['curator']].add(cur['source_hash'])
                        curators_stmt[cur['curator']].add(cur['pa_hash'])
                        curs_by_tags[cur['tag']] += 1
                        cur_ev_dates[cur['date'].strftime(df)].add(
                            cur['source_hash'])
                        cur_stmt_dates[cur['date'].strftime(df)].add(
                            cur['pa_hash'])
        for cur, entries in curators_ev.items():
            curators_ev_counts[cur] = len(entries)
        for cur, entries in curators_stmt.items():
            curators_stmt_counts[cur] = len(entries)
        current_ev_sum = 0
        current_stmt_sum = 0
        for date, entries in sorted(cur_ev_dates.items()):
            current_ev_sum += len(entries)
            cur_ev_date_sum.append((date, current_ev_sum))
        for date, entries in sorted(cur_stmt_dates.items()):
            current_stmt_sum += len(entries)
            cur_stmt_date_sum.append((date, current_stmt_sum))

        cur_stats = {
            'curators_ev_counts': sorted(
                curators_ev_counts.items(), key=lambda x: x[1], reverse=True),
            'curators_stmt_counts': sorted(
                curators_stmt_counts.items(), key=lambda x: x[1], reverse=True),
            'curs_by_tags': sorted(
                curs_by_tags.items(), key=lambda x: x[1], reverse=True),
            'cur_ev_dates': cur_ev_date_sum,
            'cur_stmt_dates': cur_stmt_date_sum
        }
        return cur_stats


class TestRound(Round):
    """Analyzes the results of one test round.

    Parameters
    ----------
    json_results : list[dict]
        A list of JSON formatted dictionaries to store information about the
        test results. The first dictionary contains information about the
        model. Each consecutive dictionary contains information about a single
        test applied to the model and test results.
    date_str : str
        Time when ModelManager responsible for this round was created.

    Attributes
    ----------
    mc_types_results : dict
        A dictionary mapping a type of a ModelChecker to a list of test
        results generated by this ModelChecker
    tests : list[indra.statements.Statement]
        A list of INDRA Statements used to make EMMAA tests.
    english_test_results : dict
        A dictionary mapping a test hash and a list containing its English
        description, result in Pass/Fail/n_a form and either a path if it
        was found or a result code if it was not.
    """
    def __init__(self, json_results, date_str):
        super().__init__(date_str)
        self.json_results = json_results
        mc_types = self.json_results[0].get('mc_types', ['pysb'])
        self.mc_types_results = {}
        for mc_type in mc_types:
            self.mc_types_results[mc_type] = self._get_results(mc_type)
        self.tests = self._get_tests()
        self.english_test_results = self._get_applied_tests_results()

    @classmethod
    def load_from_s3_key(cls, key, bucket=EMMAA_BUCKET_NAME):
        logger.info(f'Loading json from {key}')
        json_results = load_json_from_s3(bucket, key)
        date_str = json_results[0].get('date_str', strip_out_date(key))
        return cls(json_results, date_str)

    def get_applied_test_hashes(self):
        """Return a list of hashes for all applied tests."""
        return list(self.english_test_results.keys())

    def get_passed_test_hashes(self, mc_type='pysb'):
        """Return a list of hashes for passed tests."""
        return [test_hash for test_hash in self.english_test_results.keys() if
                self.english_test_results[test_hash][mc_type][0] == 'Pass']

    def get_total_applied_tests(self):
        """Return a number of all applied tests."""
        total = len(self.tests)
        logger.info(f'{total} tests were applied.')
        return total

    def get_number_passed_tests(self, mc_type='pysb'):
        """Return a number of all passed tests."""
        total = len(self.get_passed_test_hashes(mc_type))
        logger.info(f'{total} tests passed.')
        return total

    def passed_over_total(self, mc_type='pysb'):
        """Return a ratio of passed over total tests."""
        return self.get_number_passed_tests(mc_type)/self.get_total_applied_tests()

    def _get_applied_tests_results(self):
        """Return a dictionary mapping a test hash and a list containing its
        English description, result in Pass/Fail form and either a path if it
        was found or a result code if it was not."""
        tests_by_hash = {}
        logger.info('Retrieving test hashes, english tests and test results.')

        def get_pass_fail(res):
            # Here use result.path_found because we care if the path was found
            # and do not care about path length
            if res.path_found:
                return 'Pass'
            elif res.result_code == 'STATEMENT_TYPE_NOT_HANDLED':
                return 'n_a'
            else:
                return 'Fail'

        def get_path_or_code(ix, res, mc_type):
            path_or_code = None
            # Here use result.paths because we care about actual path (i.e.
            # we can't get a path exceeding max path length)
            if res.paths:
                try:
                    path_or_code = (
                        self.json_results[ix+1][mc_type]['path_json'])
                # if json doesn't contain some of the fields
                except KeyError:
                    pass
            # If path wasn't found or presented in json
            if not path_or_code:
                try:
                    path_or_code = (
                        self.json_results[ix+1][mc_type]['result_code'])
                except KeyError:
                    pass
            # Couldn't get either path or code description from json
            if not path_or_code:
                path_or_code = res.result_code
            return path_or_code

        for ix, test in enumerate(self.tests):
            test_hash = str(test.get_hash(refresh=True))
            tests_by_hash[test_hash] = {
                'test': self.get_english_statement(test)}
            for mc_type in self.mc_types_results:
                result = self.mc_types_results[mc_type][ix]
                tests_by_hash[test_hash][mc_type] = [
                        get_pass_fail(result),
                        get_path_or_code(ix, result, mc_type)]
        return tests_by_hash

    def get_path_stmt_counts(self):
        path_stmt_counts = self.json_results[0].get('path_stmt_counts')
        if path_stmt_counts:
            return sorted(
                path_stmt_counts.items(), key=lambda x: x[1], reverse=True)
        return []

    def _get_results(self, mc_type):
        unpickler = jsonpickle.unpickler.Unpickler()
        test_results = [unpickler.restore(result[mc_type]['result_json'])
                        for result in self.json_results[1:]]
        return test_results

    def _get_tests(self):
        tests = [Statement._from_json(res['test_json'])
                 for res in self.json_results[1:]]
        return tests


class StatsGenerator(object):
    """Parent class for classes generating statistic for a given round of
    tests or model update.

    Parameters
    ----------
    model_name : str
        A name of a model the tests were run against.
    latest_round : ModelRound or TestRound or None
        An instance of a ModelRound or TestRound to generate statistics for.
        If not given, will be generated by loading json from s3.
    previous_round : ModelRound or TestRound or None
        A different instance of a ModelRound or TestRound to find delta
        between two rounds. If not given, will be generated by loading json
        from s3.
    previous_json_stats : dict
        A JSON-formatted dictionary containing model or test statistics for
        the previous round.
    Attributes
    ----------
    json_stats : dict
        A JSON-formatted dictionary containing model or test statistics.
    """

    def __init__(self, model_name, latest_round=None, previous_round=None,
                 previous_json_stats=None, bucket=EMMAA_BUCKET_NAME):
        self.model_name = model_name
        self.bucket = bucket
        self.previous_date_str = None
        if not latest_round:
            self.latest_round = self._get_latest_round()
        else:
            self.latest_round = latest_round
        if not previous_json_stats:
            self.previous_json_stats = self._get_previous_json_stats()
        else:
            self.previous_json_stats = previous_json_stats
        if not previous_round:
            self.previous_round = self._get_previous_round()
        else:
            self.previous_round = previous_round
        self.json_stats = {}

    def make_changes_over_time(self):
        """Add changes to model and tests over time to json_stats."""
        raise NotImplementedError("Method must be implemented in child class.")

    def get_over_time(self, section, metrics, **kwargs):
        raise NotImplementedError("Method must be implemented in child class.")

    def get_dates(self):
        if not self.previous_json_stats:
            previous_dates = []
        else:
            previous_dates = (
                self.previous_json_stats['changes_over_time']['dates'])
        previous_dates.append(self.latest_round.date_str)
        return previous_dates

    def save_to_s3_key(self, stats_key):
        if self.json_stats:
            logger.info(f'Uploading statistics to {stats_key}')
            save_json_to_s3(self.json_stats, self.bucket, stats_key)

    def save_to_s3(self):
        raise NotImplementedError("Method must be implemented in child class.")

    def _get_latest_round(self):
        raise NotImplementedError("Method must be implemented in child class.")

    def _get_previous_round(self):
        raise NotImplementedError("Method must be implemented in child class.")

    def _get_previous_json_stats(self):
        raise NotImplementedError("Method must be implemented in child class.")


class ModelStatsGenerator(StatsGenerator):
    """Generates statistic for a given model update round.

    Parameters
    ----------
    model_name : str
        A name of a model the tests were run against.
    latest_round : emmaa.analyze_tests_results.ModelRound
        An instance of a ModelRound to generate statistics for. If not given,
        will be generated by loading model data from s3.
    previous_round : emmaa.analyze_tests_results.ModelRound
        A different instance of a ModelRound to find delta between two rounds.
        If not given, will be generated by loading model data from s3.
    previous_json_stats : list[dict]
        A JSON-formatted dictionary containing model statistics for previous
        update round.

    Attributes
    ----------
    json_stats : dict
        A JSON-formatted dictionary containing model statistics.
    """

    def __init__(self, model_name, latest_round=None, previous_round=None,
                 previous_json_stats=None, bucket=EMMAA_BUCKET_NAME):
        super().__init__(model_name, latest_round, previous_round,
                         previous_json_stats, bucket)

    def make_stats(self):
        """Check if two latest model rounds were found and add statistics to
        json_stats dictionary. If both latest round and previous round
        were passed or found on s3, a dictionary will have three key-value
        pairs: model_summary, model_delta, and changes_over_time.
        """
        if not self.latest_round:
            logger.info(f'Latest round for {self.model_name} is not found.')
            return
        if self.previous_json_stats and not self.previous_round:
            logger.info(f'Latest stats are found but latest round is not.')
            return
        logger.info(f'Generating stats for {self.model_name}.')
        self.make_model_summary()
        self.make_model_delta()
        self.make_paper_delta()
        self.make_paper_summary()
        self.make_curation_summary()
        self.make_changes_over_time()

    def make_model_summary(self):
        """Add latest model state summary to json_stats."""
        logger.info(f'Generating model summary for {self.model_name}.')
        self.json_stats['model_summary'] = {
            'model_name': self.model_name,
            'number_of_statements': self.latest_round.get_total_statements(),
            'stmts_type_distr': self.latest_round.get_statement_types(),
            'agent_distr': self.latest_round.get_agent_distribution(),
            'stmts_by_evidence': self.latest_round.get_statements_by_evidence(),
            'sources': self.latest_round.get_sources_distribution(),
            'all_stmts': self.latest_round.get_english_statements_by_hash()
        }

    def make_model_delta(self):
        """Add model delta between two latest model states to json_stats."""
        logger.info(f'Generating model delta for {self.model_name}.')
        if not self.previous_round:
            self.json_stats['model_delta'] = {
                'statements_hashes_delta': {'added': [], 'removed': []}}
        else:
            stmts_delta = self.latest_round.find_delta_hashes(
                self.previous_round, 'statements')
            self.json_stats['model_delta'] = {
                'statements_hashes_delta': stmts_delta}
            msg = _make_twitter_msg(self.model_name, 'stmts', stmts_delta,
                                    self.latest_round.date_str[:10])
            if msg:
                logger.info(msg)

    def make_paper_summary(self):
        """Add latest paper summary to json_stats."""
        logger.info(f'Generating model summary for {self.model_name}.')
        self.json_stats['paper_summary'] = {
            'raw_paper_ids': self.latest_round.get_all_raw_paper_ids(),
            'number_of_raw_papers': self.latest_round.get_number_raw_papers(),
            'assembled_paper_ids': (
                self.latest_round.get_all_assembled_paper_ids()),
            'number_of_assembled_papers': (
                self.latest_round.get_number_assembled_papers()),
            'stmts_by_paper': self.latest_round.stmts_by_papers,
            'paper_distr': self.latest_round.get_papers_distribution(),
            'raw_paper_counts': self.latest_round.get_raw_paper_counts()
        }
        freq_trids = [pair[0] for pair in
                      self.json_stats['paper_summary']['paper_distr'][:10]]
        new_trids = self.json_stats['paper_delta']['raw_paper_ids_delta'][
            'added']
        trids = list(set(freq_trids).union(set(new_trids)))
        titles, links = self.latest_round.get_paper_titles_and_links(trids)
        self.json_stats['paper_summary']['paper_titles'] = titles
        self.json_stats['paper_summary']['paper_links'] = links

    def make_paper_delta(self):
        """Add paper delta between two latest model states to json_stats."""
        logger.info(f'Generating paper delta for {self.model_name}.')
        if not self.previous_round or not self.previous_round.paper_ids:
            self.json_stats['paper_delta'] = {
                'raw_paper_ids_delta': {'added': [], 'removed': []},
                'assembled_paper_ids_delta': {'added': [], 'removed': []}}
        else:
            raw_paper_delta = self.latest_round.find_delta_hashes(
                self.previous_round, 'raw_papers')
            assembled_paper_delta = self.latest_round.find_delta_hashes(
                self.previous_round, 'assembled_papers')
            self.json_stats['paper_delta'] = {
                'raw_paper_ids_delta': raw_paper_delta,
                'assembled_paper_ids_delta': assembled_paper_delta}
            logger.info(f'Read {len(raw_paper_delta["added"])} new papers.')
            logger.info(f'Got assembled statements from '
                        f'{len(assembled_paper_delta["added"])} new papers.')

    def make_curation_summary(self):
        """Add latest curation summary to json_stats."""
        logger.info(f'Generating curation summary for { self.model_name}.')
        cur_stats = self.latest_round.get_curation_stats()
        self.json_stats['curation_summary'] = cur_stats

    def make_changes_over_time(self):
        """Add changes to model over time to json_stats."""
        logger.info(f'Comparing changes over time for {self.model_name}.')
        self.json_stats['changes_over_time'] = {
            'number_of_statements': self.get_over_time(
                'model_summary', 'number_of_statements'),
            'number_of_raw_papers': self.get_over_time(
                'paper_summary', 'number_of_raw_papers'),
            'number_of_assembled_papers': self.get_over_time(
                'paper_summary', 'number_of_assembled_papers'),
            'dates': self.get_dates()}

    def get_over_time(self, section, metrics, mc_type='pysb'):
        logger.info(f'Getting changes over time in {metrics} '
                    f'for {self.model_name}.')
        # First available stats
        if not self.previous_json_stats:
            previous_data = []
        else:
            previous_data = (
                self.previous_json_stats['changes_over_time'].get(metrics, []))
        previous_data.append(self.json_stats[section][metrics])
        return previous_data

    def save_to_s3(self):
        date_str = self.latest_round.date_str
        stats_key = (
            f'model_stats/{self.model_name}/model_stats_{date_str}.json')
        super().save_to_s3_key(stats_key)

    def _get_latest_round(self):
        latest_key = find_latest_s3_file(
            self.bucket, f'results/{self.model_name}/model_manager_',
            extension='.pkl')
        if latest_key is None:
            logger.info(f'Could not find a key to the latest model manager '
                        f'for {self.model_name} model.')
            return
        logger.info(f'Loading latest round from {latest_key}')
        mr = ModelRound.load_from_s3_key(latest_key, bucket=self.bucket,
                                         load_estmts=True)
        return mr

    def _get_previous_round(self):
        if not self.previous_json_stats:
            logger.info('Not loading previous round without previous stats')
            return
        previous_key = (f'results/{self.model_name}/model_manager_'
                        f'{self.previous_date_str}.pkl')
        if previous_key is None:
            logger.info(f'Could not find a key to the previous model manager '
                        f'for {self.model_name} model.')
            return
        logger.info(f'Loading previous round from {previous_key}')
        mr = ModelRound.load_from_s3_key(previous_key, bucket=self.bucket)
        return mr

    def _get_previous_json_stats(self):
        key = find_latest_s3_file(
            self.bucket, f'model_stats/{self.model_name}/model_stats_', '.json')
        # This is the first time statistics is generated for this model
        if key is None:
            logger.info(f'Could not find a key to the previous statistics ')
            return
        # If stats for this date exists, previous stats is the second latest
        if strip_out_date(key) == self.latest_round.date_str:
            logger.info(f'Statistics for latest round already exists')
            key = find_nth_latest_s3_file(
                1, self.bucket, f'model_stats/{self.model_name}/model_stats_',
                '.json')
        # Store the date string to find previous round with it
        self.previous_date_str = strip_out_date(key)
        logger.info(f'Loading earlier statistics from {key}')
        previous_json_stats = load_json_from_s3(self.bucket, key)
        return previous_json_stats


class TestStatsGenerator(StatsGenerator):
    """Generates statistic for a given test round.

    Parameters
    ----------
    model_name : str
        A name of a model the tests were run against.
    test_corpus_str : str
        A name of a test corpus the model was tested against.
    latest_round : emmaa.analyze_tests_results.TestRound
        An instance of a TestRound to generate statistics for. If not given,
        will be generated by loading test results from s3.
    previous_round : emmaa.analyze_tests_results.TestRound
        A different instance of a TestRound to find delta between two rounds.
        If not given, will be generated by loading test results from s3.
    previous_json_stats : list[dict]
        A JSON-formatted dictionary containing test statistics for previous
        test round.

    Attributes
    ----------
    json_stats : dict
        A JSON-formatted dictionary containing test statistics.
    """

    def __init__(self, model_name, test_corpus_str='large_corpus_tests',
                 latest_round=None, previous_round=None,
                 previous_json_stats=None, bucket=EMMAA_BUCKET_NAME):
        self.test_corpus = test_corpus_str
        super().__init__(model_name, latest_round, previous_round,
                         previous_json_stats, bucket)

    def make_stats(self):
        """Check if two latest test rounds were found and add statistics to
        json_stats dictionary. If both latest round and previous round
        were passed or found on s3, a dictionary will have three key-value
        pairs: test_round_summary, tests_delta, and changes_over_time.
        """
        if not self.latest_round:
            logger.info(f'Latest round for {self.model_name} is not found.')
            return
        if self.previous_json_stats and not self.previous_round:
            logger.info(f'Latest stats are found but latest round is not.')
            return
        logger.info(f'Generating stats for {self.model_name}.')
        self.make_test_summary()
        self.make_tests_delta()
        self.make_changes_over_time()

    def make_test_summary(self):
        """Add latest test round summary to json_stats."""
        logger.info(f'Generating test summary for {self.model_name}.')
        self.json_stats['test_round_summary'] = {
            'test_data': self.latest_round.json_results[0].get('test_data'),
            'number_applied_tests': self.latest_round.get_total_applied_tests(),
            'all_test_results': self.latest_round.english_test_results,
            'path_stmt_counts': self.latest_round.get_path_stmt_counts()}
        for mc_type in self.latest_round.mc_types_results:
            self.json_stats['test_round_summary'][mc_type] = {
                'number_passed_tests': (
                    self.latest_round.get_number_passed_tests(mc_type)),
                'passed_ratio': self.latest_round.passed_over_total(mc_type)}

    def make_tests_delta(self):
        """Add tests delta between two latest test rounds to json_stats."""
        logger.info(f'Generating tests delta for {self.model_name}.')
        date = self.latest_round.date_str[:10]
        test_name = None
        test_data = self.latest_round.json_results[0].get('test_data')
        if test_data:
            test_name = test_data.get('name')
        if not self.previous_round:
            tests_delta = {
                'applied_hashes_delta': {'added': [], 'removed': []}}
        else:
            applied_delta = self.latest_round.find_delta_hashes(
                self.previous_round, 'applied_tests')
            tests_delta = {
                'applied_hashes_delta': applied_delta}
            msg = _make_twitter_msg(
                self.model_name, 'applied_tests', applied_delta, date,
                test_corpus=self.test_corpus, test_name=test_name)
            if msg:
                logger.info(msg)

        for mc_type in self.latest_round.mc_types_results:
            if not self.previous_round or mc_type not in \
                    self.previous_round.mc_types_results:
                tests_delta[mc_type] = {
                    'passed_hashes_delta': {'added': [], 'removed': []}}
            else:
                passed_delta = self.latest_round.find_delta_hashes(
                    self.previous_round, 'passed_tests', mc_type=mc_type)
                tests_delta[mc_type] = {
                    'passed_hashes_delta': passed_delta}
                msg = _make_twitter_msg(
                    self.model_name, 'passed_tests', passed_delta, date,
                    mc_type, test_corpus=self.test_corpus, test_name=test_name)
                if msg:
                    logger.info(msg)
        self.json_stats['tests_delta'] = tests_delta

    def make_changes_over_time(self):
        """Add changes to tests over time to json_stats."""
        logger.info(f'Comparing changes over time for {self.model_name}.')
        self.json_stats['changes_over_time'] = {
            'number_applied_tests': self.get_over_time(
                'test_round_summary', 'number_applied_tests'),
            'dates': self.get_dates()}
        for mc_type in self.latest_round.mc_types_results:
            self.json_stats['changes_over_time'][mc_type] = {
                'number_passed_tests': self.get_over_time(
                    'test_round_summary', 'number_passed_tests', mc_type),
                'passed_ratio': self.get_over_time(
                    'test_round_summary', 'passed_ratio', mc_type)}

    def get_over_time(self, section, metrics, mc_type='pysb'):
        logger.info(f'Getting changes over time in {metrics} '
                    f'for {self.model_name}.')
        # Not mc_type relevant data
        if metrics == 'number_applied_tests':
            # First available stats
            if not self.previous_json_stats:
                previous_data = []
            else:
                previous_data = (
                    self.previous_json_stats['changes_over_time'][metrics])
            previous_data.append(self.json_stats[section][metrics])
        # Mc_type relevant data
        else:
            # First available stats
            if not self.previous_json_stats:
                previous_data = []
            else:
                # This mc_type wasn't available in previous stats
                if mc_type not in \
                        self.previous_json_stats['changes_over_time']:
                    previous_data = []
                else:
                    previous_data = (
                        self.previous_json_stats[
                            'changes_over_time'][mc_type][metrics])
            previous_data.append(self.json_stats[section][mc_type][metrics])
        return previous_data

    def save_to_s3(self):
        date_str = self.latest_round.date_str
        stats_key = (f'stats/{self.model_name}/test_stats_{self.test_corpus}_'
                     f'{date_str}.json')
        super().save_to_s3_key(stats_key)

    def _get_latest_round(self):
        latest_key = find_latest_s3_file(
            self.bucket,
            f'results/{self.model_name}/results_{self.test_corpus}',
            extension='.json')
        if latest_key is None:
            logger.info(f'Could not find a key to the latest test results '
                        f'for {self.model_name} model.')
            return
        logger.info(f'Loading latest round from {latest_key}')
        tr = TestRound.load_from_s3_key(latest_key, bucket=self.bucket)
        return tr

    def _get_previous_round(self):
        if not self.previous_json_stats:
            logger.info('Not loading previous round without previous stats')
            return
        previous_key = (f'results/{self.model_name}/results_{self.test_corpus}'
                        f'_{self.previous_date_str}.json')
        if previous_key is None:
            logger.info(f'Could not find a key to the previous test results '
                        f'for {self.model_name} model.')
            return
        logger.info(f'Loading previous round from {previous_key}')
        tr = TestRound.load_from_s3_key(previous_key, bucket=self.bucket)
        return tr

    def _get_previous_json_stats(self):
        key = find_latest_s3_file(
            self.bucket,
            f'stats/{self.model_name}/test_stats_{self.test_corpus}_', '.json')
        # This is the first time statistics is generated for this model
        if key is None:
            logger.info(f'Could not find a key to the previous statistics ')
            return
        # If stats for this date exists, previous stats is the second latest
        if strip_out_date(key) == self.latest_round.date_str:
            logger.info(f'Statistics for latest round already exists')
            key = find_nth_latest_s3_file(
                1, self.bucket,
                f'stats/{self.model_name}/test_stats_{self.test_corpus}_',
                '.json')
        # Store the date string to find previous round with it
        self.previous_date_str = strip_out_date(key)
        logger.info(f'Loading earlier statistics from {key}')
        previous_json_stats = load_json_from_s3(self.bucket, key)
        return previous_json_stats


def generate_stats_on_s3(
        model_name, mode, test_corpus_str='large_corpus_tests',
        upload_stats=True, bucket=EMMAA_BUCKET_NAME):
    """Generate statistics for latest round of model update or tests.

    Parameters
    ----------
    model_name : str
        A name of EmmaaModel.
    mode : str
        Type of stats to generate (model or tests)
    test_corpus_str : str
        A name of a test corpus.
    upload_stats : Optional[bool]
        Whether to upload latest statistics about model and a test.
        Default: True
    """
    if mode == 'model':
        sg = ModelStatsGenerator(model_name, bucket=bucket)
    elif mode == 'tests':
        sg = TestStatsGenerator(model_name, test_corpus_str, bucket=bucket)
    else:
        raise TypeError('Mode must be either model or tests')
    sg.make_stats()
    # Optionally upload stats to S3
    if upload_stats:
        sg.save_to_s3()
    return sg


def _make_twitter_msg(model_name, msg_type, delta, date, mc_type=None,
                      test_corpus=None, test_name=None, new_papers=None):
    if len(delta['added']) == 0:
        logger.info(f'No {msg_type} delta found')
        return
    if not test_name:
        test_name = test_corpus
    plural = 's' if len(delta['added']) > 1 else ''
    if msg_type == 'stmts':
        if not new_papers:
            logger.info(f'No new papers found')
            return
        else:
            paper_plural = 's' if new_papers > 1 else ''
            msg = (f'Today I read {new_papers} new publication{paper_plural} '
                   f'and learned {len(delta["added"])} new mechanism{plural}. '
                   f'See https://emmaa.indra.bio/dashboard/{model_name}'
                   f'?tab=model&date={date}#addedStmts for more '
                   'details.')
    elif msg_type == 'applied_tests':
        msg = (f'Today I applied {len(delta["added"])} new test{plural} in '
               f'the {test_name}. See '
               f'https://emmaa.indra.bio/dashboard/{model_name}?tab=tests'
               f'&test_corpus={test_corpus}&date={date}#newAppliedTests for '
               'more details.')
    elif msg_type == 'passed_tests' and mc_type:
        msg = (f'Today I explained {len(delta["added"])} new '
               f'observation{plural} in the {test_name} with my '
               f'{TWITTER_MODEL_TYPES[mc_type]} model. See '
               f'https://emmaa.indra.bio/dashboard/{model_name}?tab=tests'
               f'&test_corpus={test_corpus}&date={date}#newPassedTests for '
               'more details.')
    else:
        raise TypeError(f'Invalid message type: {msg_type}.')
    return msg


def tweet_deltas(model_name, test_corpora, date, bucket=EMMAA_BUCKET_NAME):
    model_stats, _ = get_model_stats(model_name, 'model', date=date)
    test_stats_by_corpus = {}
    for test_corpus in test_corpora:
        test_stats, _ = get_model_stats(model_name, 'test', tests=test_corpus,
                                        date=date)
        if not test_stats:
            logger.info(f'Could not find test stats for {test_corpus}')
        test_stats_by_corpus[test_corpus] = test_stats
    if not model_stats or not test_stats_by_corpus:
        logger.warning('Stats are not found, not tweeting')
        return
    config = load_config_from_s3(model_name, bucket)
    twitter_key = config.get('twitter')
    twitter_cred = get_credentials(twitter_key)
    if not twitter_cred:
        logger.warning('Twitter credentials are not found, not tweeting')
    # Model message
    stmts_delta = model_stats['model_delta']['statements_hashes_delta']
    paper_delta = model_stats['paper_delta']['raw_paper_ids_delta']
    new_papers = len(paper_delta['added'])
    stmts_msg = _make_twitter_msg(model_name, 'stmts', stmts_delta, date,
                                  new_papers=new_papers)
    if stmts_msg:
        logger.info(stmts_msg)
        if twitter_cred:
            update_status(stmts_msg, twitter_cred)

    # Tests messages
    for test_corpus, test_stats in test_stats_by_corpus.items():
        test_name = None
        test_data = test_stats['test_round_summary'].get('test_data')
        if test_data:
            test_name = test_data.get('name')
        for k, v in test_stats['tests_delta'].items():
            if k == 'applied_hashes_delta':
                applied_delta = v
                applied_msg = _make_twitter_msg(
                    model_name, 'applied_tests', applied_delta, date,
                    test_corpus=test_corpus, test_name=test_name)
                if applied_msg:
                    logger.info(applied_msg)
                    if twitter_cred:
                        update_status(applied_msg, twitter_cred)
            else:
                mc_type = k
                passed_delta = v['passed_hashes_delta']
                passed_msg = _make_twitter_msg(
                    model_name, 'passed_tests', passed_delta,
                    date, mc_type, test_corpus=test_corpus,
                    test_name=test_name)
                if passed_msg:
                    logger.info(passed_msg)
                    if twitter_cred:
                        update_status(passed_msg, twitter_cred)

    logger.info('Done tweeting')


def _get_pmid_titles(pmids):
    pmids_to_titles = {}
    n = 200
    n_batches = len(pmids) // n
    if len(pmids) % n:
        n_batches += 1
    for i in range(n_batches):
        start = n * i
        end = start + n
        batch = pmids[start: end]
        m = pubmed_client.get_metadata_for_ids(batch)
        for pmid, metadata in m.items():
            pmids_to_titles[pmid] = metadata['title']
    return pmids_to_titles


def _get_doi_title(doi):
    m = crossref_client.get_metadata(doi)
    if m:
        title = m.get('title')
        if title:
            return title[0]


def _get_pmcid_title(pmcid):
    title = pmc_client.get_title(pmcid)
    return title


def _get_publication_link(paper_id, paper_id_type):
    if paper_id_type == 'PMID':
        name = 'PubMed'
        link = f'https://pubmed.ncbi.nlm.nih.gov/{paper_id}'
    elif paper_id_type == 'PMCID':
        name = 'PMC'
        link = f'https://www.ncbi.nlm.nih.gov/pmc/articles/{paper_id}'
    elif paper_id_type == 'DOI':
        name = 'DOI'
        link = f'https://dx.doi.org/{paper_id}'
    return (link, name)



import time
import logging
import datetime
import pybel
from botocore.exceptions import ClientError
from indra.databases import ndex_client
from indra.literature import pubmed_client, elsevier_client, biorxiv_client
from indra.assemblers.cx import CxAssembler
from indra.assemblers.pysb import PysbAssembler
from indra.assemblers.pybel import PybelAssembler
from indra.assemblers.indranet import IndraNetAssembler
from indra.statements import stmts_from_json
from indra.pipeline import AssemblyPipeline, register_pipeline
from indra.tools.assemble_corpus import filter_grounded_only
from indra_db.client.principal.curation import get_curations
from indra_db.util import get_db, _get_trids
from emmaa.priors import SearchTerm
from emmaa.readers.aws_reader import read_pmid_search_terms
from emmaa.readers.db_client_reader import read_db_pmid_search_terms, \
    read_db_doi_search_terms
from emmaa.readers.elsevier_eidos_reader import \
    read_elsevier_eidos_search_terms
from emmaa.util import make_date_str, find_latest_s3_file, strip_out_date, \
    EMMAA_BUCKET_NAME, find_nth_latest_s3_file, load_pickle_from_s3, \
    save_pickle_to_s3, load_json_from_s3, save_json_to_s3, \
    load_gzip_json_from_s3, get_s3_client
from emmaa.statements import to_emmaa_stmts


logger = logging.getLogger(__name__)
register_pipeline(get_curations)


class EmmaaModel(object):
    """Represents an EMMAA model.

    Parameters
    ----------
    name : str
        The name of the model.
    config : dict
        A configuration dict that is typically loaded from a YAML file.
    paper_ids : list(str) or None
        A list of paper IDs used to get statements for the current state of the
        model. With new reading results, new paper IDs will be added. If not
        provided, initial set will be derived from existing statements.

    Attributes
    ----------
    stmts : list[emmaa.EmmaaStatement]
        A list of EmmaaStatement objects representing the model
    assembly_config : dict
        Configurations for assembling the model.
    test_config : dict
        Configurations for running tests on the model.
    reading_config : dict
        Configurations for reading the content.
    query_config : dict
        Configurations for running queries on the model.
    search_terms : list[emmaa.priors.SearchTerm]
        A list of SearchTerm objects containing the search terms used in the
        model.
    ndex_network : str
        The identifier of the NDEx network corresponding to the model.
    assembled_stmts : list[indra.statements.Statement]
        A list of assembled INDRA Statements
    """
    def __init__(self, name, config, paper_ids=None):
        self.name = name
        self.stmts = []
        self.assembly_config = {}
        self.test_config = {}
        self.reading_config = {}
        self.query_config = {}
        self.search_terms = []
        self.ndex_network = None
        self.human_readable_name = None
        self.export_formats = []
        self._load_config(config)
        self.assembled_stmts = []
        if paper_ids:
            self.paper_ids = set(paper_ids)
        else:
            self.paper_ids = set()
        self.date_str = make_date_str()

    def add_statements(self, stmts):
        """Add a set of EMMAA Statements to the model

        Parameters
        ----------
        stmts : list[emmaa.EmmaaStatement]
            A list of EMMAA Statements to add to the model
        """
        self.stmts += stmts

    def get_indra_stmts(self):
        """Return the INDRA Statements contained in the model.

        Returns
        -------
        list[indra.statements.Statement]
            The list of INDRA Statements that are extracted from the EMMAA
            Statements.
        """
        return [es.stmt for es in self.stmts]

    def _load_config(self, config):
        self.search_terms = [SearchTerm.from_json(s) for s in
                             config['search_terms']]
        if 'ndex' in config:
            self.ndex_network = config['ndex']['network']
        else:
            self.ndex_network = None
        if 'reading' in config:
            self.reading_config = config['reading']
        if 'assembly' in config:
            self.assembly_config = config['assembly']
        if 'test' in config:
            self.test_config = config['test']
        if 'query' in config:
            self.query_config = config['query']
        if 'human_readable_name' in config:
            self.human_readable_name = config['human_readable_name']
        self.export_formats = config.get('export_formats', [])

    def search_literature(self, lit_source, date_limit=None):
        """Search for the model's search terms in the literature.

        Parameters
        ----------
        date_limit : Optional[int]
            The number of days to search back from today.

        Returns
        -------
        ids_to_terms : dict
            A dict representing all the literature source IDs (e.g.,
            PMIDs or PIIS) returned by the searches as keys,
            and the search terms for which the given ID was produced as
            values.
        """
        if lit_source == 'pubmed':
            terms_to_ids = self.search_pubmed(self.search_terms, date_limit)
        elif lit_source == 'biorxiv':
            collection_id = self.reading_config.get('collection_id', '181')
            terms_to_ids = self.search_biorxiv(collection_id, date_limit)
        elif lit_source == 'elsevier':
            terms_to_ids = self.search_elsevier(self.search_terms, date_limit)
        else:
            raise ValueError('Unknown literature source: %s' % lit_source)
        ids_to_terms = {}
        for term, ids in terms_to_ids.items():
            for id in ids:
                try:
                    ids_to_terms[id].append(term)
                except KeyError:
                    ids_to_terms[id] = [term]
        return ids_to_terms

    @staticmethod
    def search_pubmed(search_terms, date_limit):
        """Search PubMed for given search terms.

        Parameters
        ----------
        search_terms : list[emmaa.priors.SearchTerm]
            A list of SearchTerm objects to search PubMed for.
        date_limit : int
            The number of days to search back from today.

        Returns
        -------
        terms_to_pmids : dict
            A dict representing given search terms as keys and PMIDs returned
            by searches as values.
        """
        terms_to_pmids = {}
        for term in search_terms:
            pmids = pubmed_client.get_ids(term.search_term, reldate=date_limit)
            logger.info(f'{len(pmids)} PMIDs found for {term.search_term}')
            terms_to_pmids[term] = pmids
            time.sleep(1)
        return terms_to_pmids

    @staticmethod
    def search_elsevier(search_terms, date_limit):
        """Search Elsevier for given search terms.

        Parameters
        ----------
        search_terms : list[emmaa.priors.SearchTerm]
            A list of SearchTerm objects to search PubMed for.
        date_limit : int
            The number of days to search back from today.

        Returns
        -------
        terms_to_piis : dict
            A dict representing given search terms as keys and PIIs returned
            by searches as values.
        """
        start_date = (
            datetime.datetime.utcnow() - datetime.timedelta(days=date_limit))
        start_date = start_date.isoformat(timespec='seconds') + 'Z'
        terms_to_piis = {}
        for term in search_terms:
            # NOTE for now limiting the search to only 5 PIIs
            piis = elsevier_client.get_piis_for_date(
                term.search_term, loaded_after=start_date)[:5]
            logger.info(f'{len(piis)} PIIs found for {term.search_term}')
            terms_to_piis[term] = piis
            time.sleep(1)
        return terms_to_piis

    @staticmethod
    def search_biorxiv(collection_id, date_limit):
        """Search BioRxiv within date_limit.

        Parameters
        ----------
        date_limit : int
            The number of days to search back from today.
        collection_id : str
            ID of a collection to search BioArxiv for.
        Returns
        -------
        terms_to_dois : dict
            A dict representing biorxiv collection ID as key and DOIs returned
            by search as values.
        """
        start_date = (
            datetime.datetime.utcnow() - datetime.timedelta(days=date_limit))
        dois = biorxiv_client.get_collection_dois(collection_id, start_date)
        logger.info(f'{len(dois)} DOIs found')
        term = SearchTerm('other', f'biorxiv: {collection_id}', {}, None)
        terms_to_dois = {term: dois}
        return terms_to_dois

    def get_new_readings(self, date_limit=10):
        """Search new literature, read, and add to model statements"""
        readers = self.reading_config.get('reader', ['indra_db_pmid'])
        lit_sources = self.reading_config.get('literature_source', ['pubmed'])
        if isinstance(lit_sources, str):
            lit_sources = [lit_sources]
        if isinstance(readers, str):
            readers = [readers]
        estmts = []
        for lit_source, reader in zip(lit_sources, readers):
            ids_to_terms = self.search_literature(lit_source, date_limit)
            if reader == 'aws':
                new_estmts = read_pmid_search_terms(
                    ids_to_terms)
                self.add_paper_ids(ids_to_terms.keys(), 'pmid')
            elif reader == 'indra_db_pmid':
                new_estmts = read_db_pmid_search_terms(
                    ids_to_terms)
                self.add_paper_ids(ids_to_terms.keys(), 'pmid')
            elif reader == 'indra_db_doi':
                new_estmts = read_db_doi_search_terms(
                    ids_to_terms)
                self.add_paper_ids(ids_to_terms.keys(), 'doi')
            elif reader == 'elsevier_eidos':
                new_estmts = read_elsevier_eidos_search_terms(
                    ids_to_terms)
                self.add_paper_ids(ids_to_terms.keys(), 'pii')
            else:
                raise ValueError('Unknown reader: %s' % reader)
            estmts += new_estmts
        logger.info('Got a total of %d new EMMAA Statements from reading' %
                    len(estmts))
        self.extend_unique(estmts)
        if self.reading_config.get('cord19_update'):
            self.update_with_cord19()
        self.eliminate_copies()

    def extend_unique(self, estmts):
        """Extend model statements only if it is not already there."""
        source_hashes = {est.stmt.get_hash(shallow=False, refresh=True)
                         for est in self.stmts}
        len_before = len(self.stmts)
        for estmt in estmts:
            if estmt.stmt.get_hash(shallow=False, refresh=True) not in \
                    source_hashes:
                self.stmts.append(estmt)
        len_after = len(self.stmts)
        logger.info('Extended EMMAA Statements by %d new Statements' %
                    (len_after - len_before))

    def update_with_cord19(self):
        """Update model with new CORD19 dataset statements."""
        # Using local import to avoid dependency
        from covid_19.emmaa_update import make_model_stmts
        current_stmts = self.get_indra_stmts()
        default_filenames = [
            'drug_stmts_v2.pkl', 'gordon_ndex_stmts.pkl',
            'virhostnet_stmts.pkl', 'ctd_stmts.pkl']
        if isinstance(self.reading_config['cord19_update'], dict):
            fnames = self.reading_config['cord19_update'].get(
                'filenames', default_filenames)
        else:  # if it's a boolean
            fnames = default_filenames
        other_stmts = []
        for fname in fnames:
            file_stmts = load_pickle_from_s3('indra-covid19', fname)
            logger.info(f'Loaded {len(file_stmts)} statements from {fname}.')
            other_stmts += file_stmts
        new_stmts, paper_ids = make_model_stmts(current_stmts, other_stmts)
        self.stmts = to_emmaa_stmts(new_stmts, datetime.datetime.now(), [])
        self.add_paper_ids(paper_ids, 'TRID')

    def add_paper_ids(self, initial_ids, id_type='pmid'):
        """Convert if needed and save paper IDs.

        Parameters
        ----------
        initial_ids : set(str)
            A set of paper IDs.
        id_type : str
            What type the given IDs are (e.g. pmid, doi, pii). All IDs except
            for PIIs will be converted into TextRef IDs before saving.
        """
        logger.info(f'Adding new paper IDs from {len(initial_ids)} {id_type}s')
        if id_type in {'pii', 'TRID'}:
            self.paper_ids.update(set(initial_ids))
        else:
            db = get_db('primary')
            for paper_id in initial_ids:
                trids = _get_trids(db, paper_id, id_type)
                # Some papers might be not in the database yet
                if trids:
                    self.paper_ids.add(trids[0])

    def get_paper_ids_from_stmts(self, stmts):
        """Get initial set of paper IDs from a list of statements.

        Parameters
        ----------
        stmts : list[emmaa.statements.EmmaaStatement]
            A list of EMMAA statements to create the mappings from.
        """
        main_id_type = self.reading_config.get('main_id_type', 'TRID')
        logger.info(f'Extracting {main_id_type}s from statements')
        paper_ids = set()
        for estmt in stmts:
            for evid in estmt.stmt.evidence:
                if main_id_type == 'pii':
                    paper_id = evid.annotations.get('pii')
                else:
                    paper_id = evid.text_refs.get(main_id_type)
                    # In some TextRefs the keys might be lowercase
                    if not paper_id:
                        paper_id = evid.text_refs.get(main_id_type.lower())
                if paper_id:
                    paper_ids.add(paper_id)
        logger.info(f'Got {len(paper_ids)} {main_id_type}s from statements')
        return paper_ids

    def eliminate_copies(self):
        """Filter out exact copies of the same Statement."""
        logger.info('Starting with %d raw EmmaaStatements' % len(self.stmts))
        self.stmts = list({estmt.stmt.get_hash(shallow=False, refresh=True):
                           estmt for estmt in self.stmts}.values())
        logger.info(('Continuing with %d raw EmmaaStatements'
                     ' that are not exact copies') % len(self.stmts))

    def run_assembly(self):
        """Run INDRA's assembly pipeline on the Statements."""
        self.eliminate_copies()
        stmts = self.get_indra_stmts()
        stnames = {s.name for s in self.search_terms}
        ap = AssemblyPipeline(self.assembly_config)
        self.assembled_stmts = ap.run(stmts, stnames=stnames)

    def update_to_ndex(self):
        """Update assembled model as CX on NDEx, updates existing network."""
        if not self.assembled_stmts:
            self.run_assembly()
        cxa = CxAssembler(self.assembled_stmts, network_name=self.name)
        cxa.make_model()
        cx_str = cxa.print_cx()
        ndex_client.update_network(cx_str, self.ndex_network)

    def upload_to_ndex(self):
        """Upload the assembled model as CX to NDEx, creates new network."""
        if not self.assembled_stmts:
            self.run_assembly()
        cxa = CxAssembler(self.assembled_stmts, network_name=self.name)
        cxa.make_model()
        model_uuid = cxa.upload_model()
        self.ndex_network = model_uuid
        return model_uuid

    def save_to_s3(self, bucket=EMMAA_BUCKET_NAME):
        """Dump the model state to S3."""
        fname = f'models/{self.name}/model_{self.date_str}'
        # Dump as pickle
        save_pickle_to_s3(self.stmts, bucket, key=fname+'.pkl')
        # Save ids to stmt hashes mapping as json
        id_fname = f'papers/{self.name}/paper_ids_{self.date_str}.json'
        save_json_to_s3(list(self.paper_ids), bucket, key=id_fname)
        # Dump as json
        # save_json_to_s3(self.to_json(), bucket, key=fname+'.json')

    @classmethod
    def load_from_s3(klass, model_name, bucket=EMMAA_BUCKET_NAME):
        """Load the latest model state from S3.

        Parameters
        ----------
        model_name : str
            Name of model to load. This function expects the latest model
            to be found on S3 in the emmaa bucket with key
            'models/{model_name}/model_{date_string}', and the model config
            file at 'models/{model_name}/config.json'.

        Returns
        -------
        emmaa.model.EmmaaModel
            Latest instance of EmmaaModel with the given name, loaded from S3.
        """
        config = load_config_from_s3(model_name, bucket=bucket)
        stmts, stmts_key = load_stmts_from_s3(model_name, bucket=bucket)
        date = strip_out_date(stmts_key)
        # Stmts and papers should be from the same date
        key = f'papers/{model_name}/paper_ids_{date}.json'
        try:
            paper_ids = load_json_from_s3(bucket, key)
        except ClientError as e:
            logger.warning(f'Could not find paper IDs mapping due to: {e}')
            paper_ids = None
        em = klass(model_name, config, paper_ids)
        em.stmts = stmts
        if not paper_ids:
            em.paper_ids = em.get_paper_ids_from_stmts(stmts)
        return em

    def get_entities(self):
        """Return a list of Agent objects that the model contains."""
        istmts = self.get_indra_stmts()
        agents = []
        for stmt in istmts:
            agents += [a for a in stmt.agent_list() if a is not None]
        return agents

    def get_assembled_entities(self):
        """Return a list of Agent objects that the assembled model contains."""
        if not self.assembled_stmts:
            self.run_assembly()
        agents = []
        for stmt in self.assembled_stmts:
            agents += [a for a in stmt.agent_list() if a is not None]
        return agents

    def assemble_pysb(self, mode='local', bucket=EMMAA_BUCKET_NAME):
        """Assemble the model into PySB and return the assembled model."""
        if not self.assembled_stmts:
            self.run_assembly()
        pa = PysbAssembler()
        pa.add_statements(self.assembled_stmts)
        pysb_model = pa.make_model()
        if mode == 's3':
            for exp_f in self.export_formats:
                if exp_f not in {'sbml', 'kappa', 'kappa_im', 'kappa_cm',
                                 'bngl', 'sbgn', 'pysb_flat'}:
                    continue
                fname = f'{exp_f}_{self.date_str}.{exp_f}'
                pa.export_model(exp_f, fname)
                logger.info(f'Uploading {fname}')
                client = get_s3_client(unsigned=False)
                client.upload_file(fname, bucket,
                                   f'exports/{self.name}/{fname}')
        return pysb_model

    def assemble_pybel(self, mode='local', bucket=EMMAA_BUCKET_NAME):
        """Assemble the model into PyBEL and return the assembled model."""
        if not self.assembled_stmts:
            self.run_assembly()
        pba = PybelAssembler(self.assembled_stmts)
        pybel_model = pba.make_model()
        if mode == 's3' and 'pybel' in self.export_formats:
            fname = f'pybel_{self.date_str}.bel.nodelink.json.gz'
            pybel.dump(pybel_model, fname)
            logger.info(f'Uploading {fname}')
            client = get_s3_client(unsigned=False)
            client.upload_file(fname, bucket, f'exports/{self.name}/{fname}')
        return pybel_model

    def assemble_signed_graph(self, mode='local', bucket=EMMAA_BUCKET_NAME):
        """Assemble the model into signed graph and return the assembled graph.
        """
        if not self.assembled_stmts:
            self.run_assembly()
        ia = IndraNetAssembler(self.assembled_stmts)
        signed_graph = ia.make_model(graph_type='signed')
        if mode == 's3' and 'indranet' in self.export_formats:
            fname = f'indranet_{self.date_str}.tsv'
            df = ia.make_df()
            df.to_csv(fname, sep='\t', index=False)
            logger.info(f'Uploading {fname}')
            client = get_s3_client(unsigned=False)
            client.upload_file(fname, bucket, f'exports/{self.name}/{fname}')
        return signed_graph

    def assemble_unsigned_graph(self, **kwargs):
        """Assemble the model into unsigned graph and return the assembled
        graph."""
        if not self.assembled_stmts:
            self.run_assembly()
        ia = IndraNetAssembler(self.assembled_stmts)
        unsigned_graph = ia.make_model(graph_type='digraph')
        return unsigned_graph

    def to_json(self):
        """Convert the model into a json dumpable dictionary"""
        logger.info('Converting a model to JSON')
        json_output = {'name': self.name,
                       'ndex_network': self.ndex_network,
                       'search_terms': [st.to_json() for st
                                        in self.search_terms],
                       'stmts': [st.to_json() for st in self.stmts]}
        return json_output

    def __repr__(self):
        return "EmmaModel(%s, %d stmts, %d search terms)" % \
                   (self.name, len(self.stmts), len(self.search_terms))


@register_pipeline
def filter_relevance(stmts, stnames, policy=None):
    """Filter a list of Statements to ones matching a search term."""
    logger.info('Filtering %d statements for relevance...' % len(stmts))
    stmts_out = []
    for stmt in stmts:
        agnames = {a.name for a in stmt.agent_list() if a is not None}
        if policy == 'prior_one' and (agnames & stnames):
            stmts_out.append(stmt)
        elif policy == 'prior_all' and agnames.issubset(stnames):
            stmts_out.append(stmt)
        elif policy is None:
            stmts_out.append(stmt)
    logger.info('%d statements after filter...' % len(stmts_out))
    return stmts_out


@register_pipeline
def filter_eidos_ungrounded(stmts):
    """Filter out statements from Eidos with ungrounded agents."""
    logger.info(
        'Filtering out ungrounded Eidos statements from %d statements...'
        % len(stmts))
    stmts_out = []
    eidos_stmts = []
    for stmt in stmts:
        if stmt.evidence[0].source_api == 'eidos':
            eidos_stmts.append(stmt)
        else:
            stmts_out.append(stmt)
    eidos_grounded = filter_grounded_only(eidos_stmts)
    stmts_out += eidos_grounded
    logger.info('%d statements after filter...' % len(stmts_out))
    return stmts_out


def load_config_from_s3(model_name, bucket=EMMAA_BUCKET_NAME):
    """Return a JSON dict of config settings for a model from S3.

    Parameters
    ----------
    model_name : str
        The name of the model whose config should be loaded.

    Returns
    -------
    config : dict
        A JSON dictionary of the model configuration loaded from S3.
    """
    base_key = f'models/{model_name}'
    config_key = f'{base_key}/config.json'
    logger.info(f'Loading model config from {config_key}')
    config = load_json_from_s3(bucket, config_key)
    return config


def save_config_to_s3(model_name, config, bucket=EMMAA_BUCKET_NAME):
    """Upload config settings for a model to S3.

    Parameters
    ----------
    model_name : str
        The name of the model whose config should be saved to S3.
    config : dict
        A JSON dict of configurations for the model.
    """
    base_key = f'models/{model_name}'
    config_key = f'{base_key}/config.json'
    logger.info(f'Saving model config to {config_key}')
    save_json_to_s3(config, bucket, config_key)


def load_stmts_from_s3(model_name, bucket=EMMAA_BUCKET_NAME):
    """Return the list of EMMAA Statements constituting the latest model.

    Parameters
    ----------
    model_name : str
        The name of the model whose config should be loaded.

    Returns
    -------
    stmts : list of emmaa.statements.EmmaaStatement
        The list of EMMAA Statements in the latest model version.
    """
    base_key = f'models/{model_name}'
    latest_model_key = find_latest_s3_file(bucket, f'{base_key}/model_',
                                           extension='.pkl')
    logger.info(f'Loading model state from {latest_model_key}')
    stmts = load_pickle_from_s3(bucket, latest_model_key)
    return stmts, latest_model_key


def _default_test(model, config=None, bucket=EMMAA_BUCKET_NAME):
    if config:
        return config['test']['default_test_corpus']
    else:
        config = load_config_from_s3(model, bucket=bucket)
        return _default_test(model, config, bucket=bucket)


def last_updated_date(model, file_type='model', date_format='date',
                      tests='large_corpus_tests', extension='.pkl', n=0,
                      bucket=EMMAA_BUCKET_NAME):
    """Find the most recent or the nth file of given type on S3 and return its
    creation date.

    Example file name:
    models/aml/model_2018-12-13-18-11-54.pkl

    Parameters
    ----------
    model : str
        Model name to look for
    file_type : str
        Type of a file to find the latest file for. Accepted values: 'model',
        'test_results', 'model_stats', 'test_stats'.
    date_format : str
        Format of the returned date. Accepted values are 'datetime' (returns a
        date in the format "YYYY-MM-DD-HH-mm-ss") and 'date' (returns a date
        in the format "YYYY-MM-DD"). Default is 'date'.
    extension : str
        The extension the model file needs to have. Default is '.pkl'
    n : int
        Index of the file in list of S3 files sorted by date (0-indexed).
    bucket : str
        Name of bucket on S3.

    Returns
    -------
    last_updated : str
        A string of the selected format.
    """
    if file_type == 'model':
        folder_name = 'models'
        prefix_new = prefix_old = f'models/{model}/model_'
    elif file_type == 'test_results':
        prefix_new = f'results/{model}/results_{tests}'
        prefix_old = f'results/{model}/results_'
    elif file_type == 'model_stats':
        prefix_new = f'model_stats/{model}/model_stats_'
        prefix_old = f'stats/{model}/stats_'
    elif file_type == 'test_stats':
        prefix_new = f'stats/{model}/test_stats_{tests}'
        prefix_old = f'stats/{model}/stats_'
    else:
        raise TypeError(f'Files of type {file_type} are not supported')
    try:
        return strip_out_date(
            find_nth_latest_s3_file(
                n=n,
                bucket=bucket,
                prefix=prefix_new,
                extension=extension),
            date_format=date_format)
    except TypeError:
        try:
            return strip_out_date(
                find_nth_latest_s3_file(
                    n=n,
                    bucket=bucket,
                    prefix=prefix_old,
                    extension=extension),
                date_format=date_format)
        except TypeError:
            logger.info('Could not find latest update date')
            return ''


def get_model_stats(model, mode, tests=None, date=None,
                    extension='.json', n=0, bucket=EMMAA_BUCKET_NAME):
    """Gets the latest statistics for the given model

    Parameters
    ----------
    model : str
        Model name to look for
    mode : str
        Type of stats to generate (model or test)
    tests : str
        A name of a test corpus. Default is large_corpus_tests.
    date : str or None
        Date for which the stats will be returned in "YYYY-MM-DD" format.
    extension : str
        Extension of the file.
    n : int
        Index of the file in list of S3 files sorted by date (0-indexed).
    bucket : str
        Name of bucket on S3.
    Returns
    -------
    model_data : json
        The json formatted data containing the statistics for the model
    """
    if not tests:
        tests = _default_test(model, bucket=bucket)
    # If date is not specified, get the latest or the nth
    if not date:
        if mode == 'model':
            date = last_updated_date(model, 'model_stats', 'date',
                                     extension=extension, n=n, bucket=bucket)
        elif mode == 'test':
            date = last_updated_date(model, 'test_stats', 'date', tests=tests,
                                     extension=extension, n=n, bucket=bucket)
        else:
            raise TypeError('Mode must be either model or tests')

    # Try find new formatted stats (separate for model and tests)
    if mode == 'model':
        # File name example:
        # model_stats/skcm/model_stats_2019-08-20-17-34-40.json
        prefix = f'model_stats/{model}/model_stats_{date}'
        latest_file_key = find_latest_s3_file(bucket=bucket,
                                              prefix=prefix,
                                              extension=extension)
    elif mode == 'test':
        # File name example:
        # stats/skcm/test_stats_large_corpus_tests_2019-08-20-17-34-40.json
        prefix = f'stats/{model}/test_stats_{tests}_{date}'
        latest_file_key = find_latest_s3_file(bucket=bucket,
                                              prefix=prefix,
                                              extension=extension)
    else:
        raise TypeError('Mode must be either model or tests')
    # This might be an older file with model and test stats combined.
    # File name example: stats/skcm/stats_2019-08-20-17-34-40.json
    if not latest_file_key and (
        mode == 'model' or (
            mode == 'test' and tests == _default_test(model))):
        prefix = f'stats/{model}/stats_{date}'
        latest_file_key = find_latest_s3_file(bucket=bucket,
                                              prefix=prefix,
                                              extension=extension)
    # If we still didn't filnd the file it probably does not exist
    if not latest_file_key:
        return None, None
    return (load_json_from_s3(bucket, latest_file_key),
            latest_file_key)


def get_assembled_statements(model, date=None, bucket=EMMAA_BUCKET_NAME):
    """Load and return a list of assembled statements.

    Parameters
    ----------
    model : str
        A name of a model.
    date : str or None
        Date in "YYYY-MM-DD" format for which to load the statements. If None,
        loads the latest available statements.
    bucket : str
        Name of S3 bucket to look for a file. Defaults to 'emmaa'.

    Returns
    -------
    stmts : list[indra.statements.Statement]
        A list of assembled statements.
    latest_file_key : str
        Key of a file with statements on s3.
    """
    if not date:
        prefix = f'assembled/{model}/statements_'
    else:
        prefix = f'assembled/{model}/statements_{date}'
    # Try loading gzip file
    latest_file_key = find_latest_s3_file(bucket, prefix, '.gz')
    if not latest_file_key:
        # Could be saved with .zip extension
        latest_file_key = find_latest_s3_file(bucket, prefix, '.zip')
    if latest_file_key:
        stmt_jsons = load_gzip_json_from_s3(bucket, latest_file_key)
    else:
        # Try loading json file
        latest_file_key = find_latest_s3_file(bucket, prefix, '.json')
        if latest_file_key:
            stmt_jsons = load_json_from_s3(bucket, latest_file_key)
        # Didn't get gzip, zip or json
        else:
            logger.info(f'No assembled statements found for {model}.')
            return None, None
    stmts = stmts_from_json(stmt_jsons)
    return stmts, latest_file_key


@register_pipeline
def load_custom_grounding_map(model, bucket=EMMAA_BUCKET_NAME):
    key = f'models/{model}/grounding_map.json'
    gr_map = load_json_from_s3(bucket, key)
    return gr_map



import os
import json
from copy import deepcopy
from nose.plugins.attrib import attr

from indra.statements import Activation, ActivityCondition, Phosphorylation, \
    Agent, Evidence
from emmaa.analyze_tests_results import ModelRound, TestRound, \
    ModelStatsGenerator, TestStatsGenerator, _make_twitter_msg


TestRound.__test__ = False
TestStatsGenerator.__test__ = False


path_here = os.path.abspath(os.path.dirname(__file__))
previous_results_file = os.path.join(path_here, 'previous_results.json')
new_results_file = os.path.join(path_here, 'new_results.json')
previous_test_stats_file = os.path.join(path_here, 'previous_stats.json')
previous_model_stats_file = os.path.join(path_here, 'previous_model_stats.json')
with open(previous_results_file, 'r') as f:
    previous_results = json.load(f)
with open(new_results_file, 'r') as f:
    new_results = json.load(f)
with open(previous_test_stats_file, 'r') as f:
    previous_test_stats = json.load(f)
with open(previous_model_stats_file, 'r') as f:
    previous_model_stats = json.load(f)


previous_stmts = [
        Activation(Agent('BRAF', db_refs={'HGNC': '20974'}), Agent('MAP2K1'),
                   evidence=[Evidence(text='BRAF activates MAP2K1.',
                                      source_api='assertion',
                                      text_refs={'TRID': '1234'})]),
        Activation(Agent('MAP2K1', activity=ActivityCondition('activity', True)),
                   Agent('MAPK1'),
                   evidence=[Evidence(text='Active MAP2K1 activates MAPK1.',
                                      source_api='assertion',
                                      text_refs={'TRID': '1234'})])
        ]

new_stmts = previous_stmts + [
    Activation(Agent('BRAF', db_refs={'HGNC': '1097'}),
               Agent('AKT', db_refs={'FPLX': 'AKT'}),
               evidence=[Evidence(text='BRAF activates AKT',
                                  source_api='test_source1',
                                  text_refs={'TRID': '2345'})]),
    Activation(Agent('AKT', db_refs={'FPLX': 'AKT'},
                     activity=ActivityCondition('activity', True)),
               Agent('MTOR', db_refs={"HGNC": "3942"}),
               evidence=[Evidence(text='AKT activate MTOR',
                                  source_api='test_source2',
                                  text_refs={'TRID': '2345'})])]


previous_papers = {'1234'}
new_papers = {'1234', '2345', '3456'}


def test_model_round():
    mr = ModelRound(previous_stmts, '2020-01-01-00-00-00', previous_papers)
    assert mr
    assert mr.get_total_statements() == 2
    assert len(mr.get_stmt_hashes()) == 2
    assert mr.get_statement_types() == [('Activation', 2)]
    assert all(agent_tuple in mr.get_agent_distribution() for agent_tuple in
               [('BRAF', 1), ('MAP2K1', 2), ('MAPK1', 1)])
    assert all((stmt_hash, 1) in mr.get_statements_by_evidence() for stmt_hash
               in mr.get_stmt_hashes())
    assert mr.get_sources_distribution() == [('assertion', 2)]
    assert mr.get_number_raw_papers() == 1, mr.get_number_raw_papers()
    assert mr.get_number_assembled_papers() == 1
    mr2 = ModelRound(new_stmts, '2020-01-02-00-00-00', new_papers)
    assert mr2
    assert mr2.get_total_statements() == 4
    assert len(mr2.get_stmt_hashes()) == 4
    assert mr2.get_statement_types() == [('Activation', 4)]
    assert all(agent_tuple in mr2.get_agent_distribution() for agent_tuple in
               [('BRAF', 2), ('MAP2K1', 2), ('MAPK1', 1), ('MTOR', 1),
                ('AKT', 2)])
    assert len(mr2.find_delta_hashes(mr, 'statements')['added']) == 2
    assert all(source_tuple in mr2.get_sources_distribution() for source_tuple
               in [('assertion', 2), ('test_source1', 1), ('test_source2', 1)])
    assert mr2.get_number_raw_papers() == 3
    assert set(mr2.find_delta_hashes(mr, 'raw_papers')['added']) == {
        '2345', '3456'}
    assert mr2.get_number_assembled_papers() == 2, mr2.stmts_by_papers
    assert set(mr2.find_delta_hashes(mr, 'assembled_papers')['added']) == {
        '2345'}


def test_test_round():
    tr = TestRound(previous_results, '2020-01-01-00-00-00')
    assert tr
    assert tr.get_total_applied_tests() == 1
    assert tr.get_number_passed_tests() == 1
    assert tr.get_applied_test_hashes() == tr.get_passed_test_hashes()
    assert tr.passed_over_total() == 1.0
    tr2 = TestRound(new_results, '2020-01-02-00-00-00')
    assert tr2
    assert tr2.get_total_applied_tests() == 2
    assert tr2.get_number_passed_tests() == 2
    assert tr2.get_applied_test_hashes() == tr2.get_passed_test_hashes()
    assert tr2.passed_over_total() == 1.0
    assert len(tr2.find_delta_hashes(tr, 'applied_tests')['added']) == 1
    assert len(tr2.find_delta_hashes(tr, 'passed_tests')['added']) == 1
    assert len(tr2.find_delta_hashes(tr, 'paths')['added']) == 1


@attr('notravis', 'nonpublic')
def test_model_stats_generator():
    latest_round = ModelRound(new_stmts, '2020-01-02-00-00-00', new_papers)
    previous_round = ModelRound(previous_stmts, '2020-01-01-00-00-00',
                                previous_papers)
    sg = ModelStatsGenerator('test', latest_round=latest_round,
                             previous_round=previous_round,
                             previous_json_stats=previous_model_stats)
    sg.make_stats()
    assert sg.json_stats
    model_summary = sg.json_stats['model_summary']
    assert model_summary['model_name'] == 'test'
    assert model_summary['number_of_statements'] == 4
    assert model_summary['stmts_type_distr'] == [('Activation', 4)]
    assert all(agent_tuple in model_summary['agent_distr'] for
               agent_tuple in [('AKT', 2), ('BRAF', 2), ('MAP2K1', 2),
                               ('MTOR', 1), ('MAPK1', 1)])
    assert all(source_tuple in model_summary['sources'] for source_tuple
               in [('assertion', 2), ('test_source1', 1), ('test_source2', 1)])
    assert len(model_summary['stmts_by_evidence']) == 4
    assert len(model_summary['all_stmts']) == 4
    model_delta = sg.json_stats['model_delta']
    assert len(model_delta['statements_hashes_delta']['added']) == 2
    paper_summary = sg.json_stats['paper_summary']
    assert set(paper_summary['raw_paper_ids']) == {'1234', '2345', '3456'}
    assert paper_summary['number_of_raw_papers'] == 3
    assert set(paper_summary['assembled_paper_ids']) == {'1234', '2345'}
    assert paper_summary['number_of_assembled_papers'] == 2
    assert all(paper_tuple in paper_summary['paper_distr'] for
               paper_tuple in [('1234', 2), ('2345', 2)])
    assert set(paper_summary['stmts_by_paper']['1234']) == {
        29000400098290256, -20714390795236750}
    assert set(paper_summary['stmts_by_paper']['2345']) == {
        2874381165909177, -13855132444206450}
    paper_delta = sg.json_stats['paper_delta']
    assert len(paper_delta['raw_paper_ids_delta']['added']) == 2
    assert len(paper_delta['assembled_paper_ids_delta']['added']) == 1
    changes = sg.json_stats['changes_over_time']
    assert changes['number_of_statements'] == [2, 4]
    assert changes['number_of_raw_papers'] == [1, 3]
    assert changes['number_of_assembled_papers'] == [1, 2]
    assert len(changes['dates']) == 2


def test_test_stats_generator():
    latest_round = TestRound(new_results, '2020-01-02-00-00-00')
    previous_round = TestRound(previous_results, '2020-01-01-00-00-00')
    sg = TestStatsGenerator('test', latest_round=latest_round,
                            previous_round=previous_round,
                            previous_json_stats=previous_test_stats)
    sg.make_stats()
    assert sg.json_stats
    test_round_summary = sg.json_stats['test_round_summary']
    assert test_round_summary['number_applied_tests'] == 2
    assert len(test_round_summary['all_test_results']) == 2
    assert test_round_summary['pysb']['number_passed_tests'] == 2
    assert test_round_summary['pysb']['passed_ratio'] == 1.0
    assert test_round_summary['pybel']['number_passed_tests'] == 2
    assert test_round_summary['pybel']['passed_ratio'] == 1.0
    assert test_round_summary['signed_graph']['number_passed_tests'] == 2
    assert test_round_summary['signed_graph']['passed_ratio'] == 1.0
    assert test_round_summary['unsigned_graph']['number_passed_tests'] == 2
    assert test_round_summary['unsigned_graph']['passed_ratio'] == 1.0
    tests_delta = sg.json_stats['tests_delta']
    assert len(tests_delta['applied_hashes_delta']['added']) == 1
    assert len(tests_delta['pysb']['passed_hashes_delta']['added']) == 1
    assert len(tests_delta['pybel']['passed_hashes_delta']['added']) == 1
    assert len(tests_delta['signed_graph']['passed_hashes_delta']['added']) == 1
    assert len(tests_delta['unsigned_graph']['passed_hashes_delta']['added']) == 1
    changes = sg.json_stats['changes_over_time']
    assert changes['number_applied_tests'] == [1, 2]
    assert len(changes['dates']) == 2
    assert changes['pysb']['number_passed_tests'] == [1, 2]
    assert changes['pysb']['passed_ratio'] == [1, 1]
    assert changes['pybel']['number_passed_tests'] == [1, 2]
    assert changes['pybel']['passed_ratio'] == [1, 1]
    assert changes['signed_graph']['number_passed_tests'] == [1, 2]
    assert changes['signed_graph']['passed_ratio'] == [1, 1]
    assert changes['unsigned_graph']['number_passed_tests'] == [1, 2]
    assert changes['unsigned_graph']['passed_ratio'] == [1, 1]


def test_twitter_msg():
    # No message when no delta
    msg = _make_twitter_msg('test', 'stmts', {'added': []}, '2020-01-01',
                            new_papers=2)
    assert not msg
    # No message with no new papers
    msg = _make_twitter_msg('test', 'stmts', {'added': [1234, 2345]},
                            '2020-01-01', new_papers=None)
    assert not msg, msg
    # New statements with new papers message
    msg = _make_twitter_msg('test', 'stmts', {'added': [1234, 2345]},
                            '2020-01-01', new_papers=5)
    assert msg == ('Today I read 5 new publications and learned 2 new '
                   'mechanisms. See '
                   'https://emmaa.indra.bio/dashboard/test?tab=model&date='
                   '2020-01-01#addedStmts for more details.'), msg
    # New applied tests message
    msg = _make_twitter_msg('test', 'applied_tests', {'added': [1234, 2345]},
                            '2020-01-01', test_corpus='simple_tests',
                            test_name='Simple tests corpus')
    assert msg == ('Today I applied 2 new tests in the Simple tests corpus. '
                   'See https://emmaa.indra.bio/dashboard/test?'
                   'tab=tests&test_corpus=simple_tests&date=2020-01-01'
                   '#newAppliedTests for more details.'), msg
    # New passed tests message
    msg = _make_twitter_msg('test', 'passed_tests', {'added': [1234, 2345]},
                            '2020-01-01', 'pysb', test_corpus='simple_tests',
                            test_name='Simple tests corpus')
    assert msg == ('Today I explained 2 new observations in the Simple tests '
                   'corpus with my @PySysBio model. See '
                   'https://emmaa.indra.bio/dashboard/test?tab=tests'
                   '&test_corpus=simple_tests&date=2020-01-01#newPassedTests '
                   'for more details.'), msg



from nose.plugins.attrib import attr

from indra.statements import *
from indra.explanation.model_checker import PathResult, PysbModelChecker, \
    PybelModelChecker, SignedGraphModelChecker, UnsignedGraphModelChecker
from emmaa.model import EmmaaModel
from emmaa.model_tests import StatementCheckingTest, ModelManager, \
    ScopeTestConnector, TestManager, RefinementTestConnector
from emmaa.analyze_tests_results import TestRound, StatsGenerator
from emmaa.tests.test_model import create_model


# Tell nose to not run tests in the imported modules
StatementCheckingTest.__test__ = False
TestRound.__test__ = False
ScopeTestConnector.__test__ = False
TestManager.__test__ = False


def test_model_manager_structure():
    model = create_model()
    mm = ModelManager(model)
    assert isinstance(mm, ModelManager)
    assert isinstance(mm.model, EmmaaModel)
    assert mm.model.name == 'test'
    assert len(mm.mc_types) == 4, len(mm.mc_types)
    assert len(mm.mc_types['pysb']) == 3
    assert isinstance(mm.mc_types['pysb']['model_checker'], PysbModelChecker)
    assert isinstance(mm.mc_types['pybel']['model_checker'], PybelModelChecker)
    assert isinstance(
        mm.mc_types['signed_graph']['model_checker'], SignedGraphModelChecker)
    assert isinstance(
        mm.mc_types['unsigned_graph']['model_checker'],
        UnsignedGraphModelChecker)
    assert isinstance(mm.entities[0], Agent)
    assert isinstance(mm.date_str, str)


def test_run_tests():
    model = create_model()
    tests = [StatementCheckingTest(
             Activation(Agent('BRAF', db_refs={'HGNC': '1097'}),
                        Agent('MAPK1', db_refs={'UP': 'P28482'})))]
    mm = ModelManager(model)
    tm = TestManager([mm], tests)
    tm.make_tests(ScopeTestConnector())
    tm.run_tests()
    assert len(mm.applicable_tests) == 1
    assert isinstance(mm.applicable_tests[0], StatementCheckingTest)
    assert len(mm.mc_types['pysb']['test_results']) == 1
    assert len(mm.mc_types['pybel']['test_results']) == 1
    assert len(mm.mc_types['signed_graph']['test_results']) == 1
    assert len(mm.mc_types['unsigned_graph']['test_results']) == 1
    assert isinstance(mm.mc_types['pysb']['test_results'][0], PathResult)


def test_applicability():
    model = create_model()
    tests = [StatementCheckingTest(
                Activation(Agent('BRAF', db_refs={'HGNC': '1097'}),
                           Agent('MAPK1', db_refs={'UP': 'P28482'}))),
             StatementCheckingTest(
                Activation(Agent('BRAF', db_refs={'HGNC': '1097'}),
                           Agent('ERK', db_refs={'FPLX': 'ERK'})))]
    mm = ModelManager(model)
    tm = TestManager([mm], tests)
    # Only first test is applicable with ScopeTestConnector
    tm.make_tests(ScopeTestConnector())
    assert len(mm.applicable_tests) == 1
    assert mm.applicable_tests[0] == tests[0]
    # Both tests are applicable with RefinementTestConnector
    mm.applicable_tests = []
    tm.make_tests(RefinementTestConnector())
    assert len(mm.applicable_tests) == 2


def test_results_json():
    model = create_model()
    model.run_assembly()
    # Add statements with similar subject and object to test grouping
    map2k1 = model.assembled_stmts[1].subj
    mapk1 = model.assembled_stmts[1].obj
    phos = Phosphorylation(map2k1, mapk1)
    phos_t185 = Phosphorylation(map2k1, mapk1, 'T', '185')
    phos_y187 = Phosphorylation(map2k1, mapk1, 'Y', '187')
    inc = IncreaseAmount(map2k1, mapk1)
    inh = Inhibition(map2k1, mapk1)
    model.assembled_stmts += [phos, phos_t185, phos_y187, inc, inh]
    mm = ModelManager(model)
    tests = [StatementCheckingTest(
                Activation(Agent('BRAF', db_refs={'HGNC': '1097'}),
                           Agent('MAPK1', db_refs={'UP': 'P28482'}))),
             StatementCheckingTest(
                 Phosphorylation(
                     Agent('MEK', db_refs={'TEXT': 'MEK', 'FPLX': 'MEK'}),
                     Agent('ERK', db_refs={'TEXT': 'ERK', 'FPLX': 'ERK'})))]
    tm = TestManager([mm], tests)
    tm.make_tests(RefinementTestConnector())
    tm.run_tests()
    result_json, json_lines = mm.results_to_json()
    assert len(result_json) == 3
    # Looking at the first result
    assert len(result_json[1]) == 6, len(result_json[1])
    # The second edge will be supported differently in different model types
    assert result_json[1]['pysb']['path_json'][0]['path'] == \
        'BRAF → MAP2K1 → MAPK1'
    second_edge = result_json[1]['pysb']['path_json'][0]['edge_list'][1]
    # Only Activation statement will be in the edge in PySB
    assert len(second_edge['stmts']) == 1
    assert second_edge['stmts'][0][1] == 'Active MAP2K1 activates MAPK1.'
    assert second_edge['stmts'][0][0].count('stmt_hash') == 1
    # Positive (Activation and IncreaseAmount) in SignedGraph edge
    assert result_json[1]['signed_graph']['path_json'][0]['path'] == \
        'BRAF → MAP2K1 → MAPK1'
    second_edge = result_json[1]['signed_graph']['path_json'][0][
        'edge_list'][1]
    assert len(second_edge['stmts']) == 2
    assert second_edge['stmts'][0][1] == 'MAP2K1 activates MAPK1.'
    assert second_edge['stmts'][0][0].count('stmt_hash') == 1
    assert second_edge['stmts'][1][1] == ('MAP2K1 increases the amount of '
                                          'MAPK1.')
    assert second_edge['stmts'][1][0].count('stmt_hash') == 1
    # All statement types support unsigned graph edge, but different statements
    # of the same type are grouped together
    assert result_json[1]['unsigned_graph']['path_json'][0]['path'] == \
        'BRAF → MAP2K1 → MAPK1'
    second_edge = result_json[1]['unsigned_graph']['path_json'][0][
        'edge_list'][1]
    assert len(second_edge['stmts']) == 4
    sentence_counts = {pair[1]: pair[0].count('stmt_hash')
                       for pair in second_edge['stmts']}
    assert 'MAP2K1 activates MAPK1.' in sentence_counts
    assert sentence_counts['MAP2K1 activates MAPK1.'] == 1
    assert 'MAP2K1 phosphorylates MAPK1.' in sentence_counts
    assert sentence_counts['MAP2K1 phosphorylates MAPK1.'] == 3
    assert 'MAP2K1 inhibits MAPK1.' in sentence_counts
    assert sentence_counts['MAP2K1 inhibits MAPK1.'] == 1
    assert 'MAP2K1 increases the amount of MAPK1.' in sentence_counts
    assert sentence_counts['MAP2K1 increases the amount of MAPK1.'] == 1
    # Test JSONL representation
    assert len(json_lines) == 6, len(json_lines)
    for path_dict in json_lines:
        # First test
        if path_dict['test'] == 13165736649758742:
            assert len(path_dict['edges']) == 2
            assert path_dict['edges'][0]['type'] == 'statements'
            assert len(path_dict['edges'][0]['hashes']) == 1
            if path_dict['graph_type'] == 'pysb':
                assert path_dict['edges'][1]['type'] == 'statements'
                assert len(path_dict['edges'][1]['hashes']) == 1
            elif path_dict['graph_type'] == 'signed_graph':
                assert path_dict['edges'][1]['type'] == 'statements'
                assert len(path_dict['edges'][1]['hashes']) == 2
            elif path_dict['graph_type'] == 'unsigned_graph':
                assert path_dict['edges'][1]['type'] == 'statements'
                assert len(path_dict['edges'][1]['hashes']) == 6
        # Second test
        else:
            if path_dict['graph_type'] == 'pysb':
                assert len(path_dict['edges']) == 3
                assert path_dict['edges'][0]['type'] == 'RefEdge'
                assert 'hashes' not in path_dict['edges'][0]
                assert path_dict['edges'][1]['type'] == 'statements'
                assert len(path_dict['edges'][1]['hashes']) == 1
                assert path_dict['edges'][2]['type'] == 'RefEdge'
                assert 'hashes' not in path_dict['edges'][2]
            elif path_dict['graph_type'] == 'pybel':
                assert len(path_dict['edges']) == 3
                assert path_dict['edges'][0]['type'] == 'RefEdge'
                assert 'hashes' not in path_dict['edges'][0]
                assert path_dict['edges'][1]['type'] == 'statements'
                assert len(path_dict['edges'][1]['hashes']) == 1
                assert path_dict['edges'][2]['type'] == 'RefEdge'
                assert 'hashes' not in path_dict['edges'][2]
            elif path_dict['graph_type'] == 'unsigned_graph':
                assert len(path_dict['edges']) == 3
                assert path_dict['edges'][0]['type'] == 'RefEdge'
                assert 'hashes' not in path_dict['edges'][0]
                assert path_dict['edges'][1]['type'] == 'statements'
                assert len(path_dict['edges'][1]['hashes']) == 6
                assert path_dict['edges'][2]['type'] == 'RefEdge'
                assert 'hashes' not in path_dict['edges'][2]



import time
import random

from nose.plugins.attrib import attr
from nose.tools import with_setup

from emmaa.db import Query, Result, User, UserQuery
from emmaa.queries import Query as QueryObject, PathProperty
from emmaa.tests.db_setup import _get_test_db, setup_function, \
    teardown_function


test_query_jsons = [{'type': 'path_property', 'path': {'type': 'Activation',
                     'subj': {'type': 'Agent', 'name': 'BRAF'},
                     'obj': {'type': 'Agent', 'name': 'ERK'}}},
                    {'type': 'path_property', 'path':
                    {'type': 'Phosphorylation',
                     'enz': {'type': 'Agent', 'name': 'ERK'},
                     'sub': {'type': 'Agent', 'name': 'MEK'}}}]

test_queries = [QueryObject._from_json(qj) for qj in test_query_jsons]


@with_setup(setup_function, teardown_function)
@attr('nonpublic')
def test_instantiation():
    db = _get_test_db()
    assert db
    return


@with_setup(setup_function, teardown_function)
@attr('nonpublic')
def test_put_queries():
    db = _get_test_db()
    db.put_queries('joshua', 1, test_queries[0], ['aml', 'luad'])
    with db.get_session() as sess:
        queries = sess.query(Query).all()
    assert len(queries) == 2, len(queries)
    # Not logged in user
    db.put_queries(None, None, test_queries[1], ['aml'])
    with db.get_session() as sess:
        queries = sess.query(Query).all()
    assert len(queries) == 3, len(queries)


@with_setup(setup_function, teardown_function)
@attr('nonpublic')
def test_get_queries():
    db = _get_test_db()
    for query in test_queries:
        db.put_queries('joshua', 1, query, ['aml', 'luad'])
        db.put_queries('test_user', 2, query, ['aml'])
    # We should only get distinct queries, independent of number of users
    queries = db.get_queries('aml')
    assert len(queries) == 2, len(queries)
    assert all(isinstance(query, PathProperty) for query in queries)
    queries = db.get_queries('luad')
    assert len(queries) == 2, len(queries)
    assert all(isinstance(query, PathProperty) for query in queries)


def _get_random_result():
    return random.choice(
        [{'12': [['This is fine.', '']]}, {'34': [['This is not ok.', '']]}])


@with_setup(setup_function, teardown_function)
@attr('nonpublic')
def test_put_results():
    db = _get_test_db()
    db.put_queries('joshua', 1, test_queries[0], ['aml', 'luad'])
    queries = db.get_queries('aml')
    results = [(query, '', _get_random_result())
               for query in queries]
    db.put_results('aml', results)
    with db.get_session() as sess:
        db_results = sess.query(Result).all()
    assert len(db_results) == len(results)


@with_setup(setup_function, teardown_function)
@attr('nonpublic')
def test_get_results():
    db = _get_test_db()
    models = ['aml', 'luad']

    # Fill up the database.
    for query in test_queries:
        db.put_queries('joshua', 1, query, models)
    for model in models:
        db.put_results(model, [(query, 'pysb', _get_random_result())
                               for query in test_queries])

    # Try to get the results.
    results = db.get_results('joshua')
    assert len(results) == len(test_queries)*len(models)
    assert all(isinstance(result, tuple) for result in results)
    assert all(result[0] in models for result in results)
    assert any(results[0][1].matches(q) for q in test_queries)
    assert all(isinstance(result[2], str) for result in results)
    assert all(isinstance(result[3], dict) for result in results)


@with_setup(setup_function, teardown_function)
@attr('nonpublic')
def test_get_latest_results():
    db = _get_test_db()
    models = ['aml', 'luad']

    # Fill up the database.
    for query in test_queries:
        db.put_queries('joshua', 1, query, models)
    for model in models:
        db.put_results(model, [(query, 'pysb', _get_random_result())
                               for query in test_queries])

    # Add the same statements over again
    time.sleep(10)
    for model in models:
        db.put_results(model, [(query, 'pysb', _get_random_result())
                               for query in test_queries])

    # Try to get the results. Make sure we only get the one set.
    results = db.get_results('joshua')
    assert len(results) == len(test_queries)*len(models), len(results)
    assert all(isinstance(result, tuple) for result in results)
    assert all(result[0] in models for result in results)
    assert any(results[0][1].matches(q) for q in test_queries)
    assert all(isinstance(result[2], str) for result in results)
    assert all(isinstance(result[3], dict) for result in results)


@with_setup(setup_function, teardown_function)
@attr('nonpublic')
def test_get_subscribed_queries():
    db = _get_test_db()
    db.put_queries('test@test.com', 1, test_queries[0], ['aml'], True)
    db.put_queries('test@test.com', 1, test_queries[1], ['aml'], False)
    # Only return queries for which subscription is True
    queries = db.get_subscribed_queries('test@test.com')
    assert len(queries) == 1
    assert queries[0][2] == test_queries[0].get_hash_with_model('aml')


@with_setup(setup_function, teardown_function)
@attr('nonpublic')
def test_get_subscribed_users():
    db = _get_test_db()
    db.put_queries('test1@test.com', 1, test_queries[0], ['aml'], True)
    db.put_queries('test2@test.com', 2, test_queries[1], ['aml'], False)
    # Only return users that subscribed for something
    emails = db.get_subscribed_users()
    assert len(emails) == 1
    assert emails[0] == 'test1@test.com'


@with_setup(setup_function, teardown_function)
@attr('nonpublic')
def test_update_email_subscription():
    db = _get_test_db()
    db.put_queries('test1@test.com', 1, test_queries[0], ['aml'], True)
    qh = test_queries[0].get_hash_with_model('aml')
    with db.get_session() as sess:
        q = sess.query(UserQuery.subscription).filter(
            UserQuery.user_id == 1, UserQuery.query_hash == qh)
    assert [q[0] for q in q.all()][0]  # True
    db.update_email_subscription('test1@test.com', [qh], False)
    with db.get_session() as sess:
        q = sess.query(UserQuery.subscription).filter(
            UserQuery.user_id == 1,
            UserQuery.query_hash == test_queries[0].get_hash_with_model('aml'))
    assert not [q[0] for q in q.all()][0]  # new subscription status is False


@with_setup(setup_function, teardown_function)
@attr('nonpublic')
def test_get_number_results():
    db = _get_test_db()
    db.put_queries('test@test.com', 1, test_queries[0], ['aml'])
    db.put_results('aml', [(test_queries[0], 'pysb', _get_random_result())])
    time.sleep(1)
    db.put_results('aml', [(test_queries[0], 'pysb', _get_random_result())])
    time.sleep(1)
    db.put_results('aml', [(test_queries[0], 'pysb', _get_random_result())])
    qh = test_queries[0].get_hash_with_model('aml')
    assert db.get_number_of_results(qh, 'pysb') == 3


@with_setup(setup_function, teardown_function)
@attr('nonpublic')
def test_get_all_result_hashes_and_delta():
    db = _get_test_db()
    db.put_queries('test@test.com', 1, test_queries[0], ['aml'])
    qh = test_queries[0].get_hash_with_model('aml')
    # If there are no older results, all hashes is None
    assert db.get_all_result_hashes(qh, 'pysb') is None
    db.put_results('aml', [(test_queries[0], 'pysb', {'1234': 'result'})])
    assert db.get_all_result_hashes(qh, 'pysb') == {'1234'}
    # First result is not delta
    results = db.get_results_from_hashes([qh])
    assert results[0][4] == [], results[0]
    # All hashes keeps growing
    time.sleep(1)
    db.put_results('aml', [(test_queries[0], 'pysb', {'345': 'other'})])
    assert db.get_all_result_hashes(qh, 'pysb') == {'1234', '345'}
    results = db.get_results_from_hashes([qh])
    assert results[0][4] == ['345'], results[0]
    # When we go to previous result, it's not delta
    time.sleep(1)
    db.put_results('aml', [(test_queries[0], 'pysb', {'1234': 'result'})])
    assert db.get_all_result_hashes(qh, 'pysb') == {'1234', '345'}
    results = db.get_results_from_hashes([qh])
    assert results[0][4] == [], results[0]



from indra.tools import assemble_corpus as ac


filter_functions = {}


def register_filter(function):
    """
    Decorator to register a function as a filter for tests.

    A function should take an agent as an argument and return True if the agent
    is allowed to be in a path and False otherwise.
    """
    filter_functions[function.__name__] = function
    return function


@register_filter
def filter_chem_mesh_go(agent):
    """Filter ungrounded agents and agents grounded to MESH, CHEBI, GO unless
    also grounded to HMDB.
    """
    gr = agent.get_grounding()
    return gr[0] not in {'MESH', 'CHEBI', 'GO', None}



from emmaa_service.api import _make_query
from emmaa.queries import PathProperty, DynamicProperty


def test_make_query():
    query, tab = _make_query({
        'typeSelection': 'Activation',
        'subjectSelection': 'BRAF',
        'objectSelection': 'MAPK1'})
    assert isinstance(query, PathProperty)
    assert tab == 'static'
    query, tab = _make_query({
        'agentSelection': 'phosphorylated MAP2K1',
        'valueSelection': 'low',
        'patternSelection': 'always_value'})    
    assert isinstance(query, DynamicProperty)
    assert tab == 'dynamic'
